# Copyright (C) 2012, 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2012 Isaku Yamahata <yamahata at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Manage switches.

Planned to be replaced by os_ken/topology.
"""

import logging
import warnings

from os_ken.base import app_manager
from os_ken.controller import event
from os_ken.controller import handler
from os_ken.controller import ofp_event
from os_ken.controller.handler import set_ev_cls
import os_ken.exception as os_ken_exc

from os_ken.lib.dpid import dpid_to_str

LOG = logging.getLogger('os_ken.controller.dpset')

DPSET_EV_DISPATCHER = "dpset"


class EventDPBase(event.EventBase):
    def __init__(self, dp):
        super(EventDPBase, self).__init__()
        self.dp = dp


class EventDP(EventDPBase):
    """
    An event class to notify connect/disconnect of a switch.

    For OpenFlow switches, one can get the same notification by observing
    os_ken.controller.ofp_event.EventOFPStateChange.
    An instance has at least the following attributes.

    ========= =================================================================
    Attribute Description
    ========= =================================================================
    dp        A os_ken.controller.controller.Datapath instance of the switch
    enter     True when the switch connected to our controller.  False for
              disconnect.
    ports     A list of port instances.
    ========= =================================================================
    """

    def __init__(self, dp, enter_leave):
        # enter_leave
        # True: dp entered
        # False: dp leaving
        super(EventDP, self).__init__(dp)
        self.enter = enter_leave
        self.ports = []  # port list when enter or leave


class EventDPReconnected(EventDPBase):
    def __init__(self, dp):
        super(EventDPReconnected, self).__init__(dp)
        # port list, which should not change across reconnects
        self.ports = []


class EventPortBase(EventDPBase):
    def __init__(self, dp, port):
        super(EventPortBase, self).__init__(dp)
        self.port = port


class EventPortAdd(EventPortBase):
    """
    An event class for switch port status "ADD" notification.

    This event is generated when a new port is added to a switch.
    For OpenFlow switches, one can get the same notification by observing
    os_ken.controller.ofp_event.EventOFPPortStatus.
    An instance has at least the following attributes.

    ========= =================================================================
    Attribute Description
    ========= =================================================================
    dp        A os_ken.controller.controller.Datapath instance of the switch
    port      port number
    ========= =================================================================
    """

    def __init__(self, dp, port):
        super(EventPortAdd, self).__init__(dp, port)


class EventPortDelete(EventPortBase):
    """
    An event class for switch port status "DELETE" notification.

    This event is generated when a port is removed from a switch.
    For OpenFlow switches, one can get the same notification by observing
    os_ken.controller.ofp_event.EventOFPPortStatus.
    An instance has at least the following attributes.

    ========= =================================================================
    Attribute Description
    ========= =================================================================
    dp        A os_ken.controller.controller.Datapath instance of the switch
    port      port number
    ========= =================================================================
    """

    def __init__(self, dp, port):
        super(EventPortDelete, self).__init__(dp, port)


class EventPortModify(EventPortBase):
    """
    An event class for switch port status "MODIFY" notification.

    This event is generated when some attribute of a port is changed.
    For OpenFlow switches, one can get the same notification by observing
    os_ken.controller.ofp_event.EventOFPPortStatus.
    An instance has at least the following attributes.

    ========= ====================================================================
    Attribute Description
    ========= ====================================================================
    dp        A os_ken.controller.controller.Datapath instance of the switch
    port      port number
    ========= ====================================================================
    """

    def __init__(self, dp, new_port):
        super(EventPortModify, self).__init__(dp, new_port)


class PortState(dict):
    def __init__(self):
        super(PortState, self).__init__()

    def add(self, port_no, port):
        self[port_no] = port

    def remove(self, port_no):
        del self[port_no]

    def modify(self, port_no, port):
        self[port_no] = port


# this depends on controller::Datapath and dispatchers in handler
class DPSet(app_manager.OSKenApp):
    """
    DPSet application manages a set of switches (datapaths)
    connected to this controller.

    Usage Example::

        # ...(snip)...
        from os_ken.controller import dpset


        class MyApp(app_manager.OSKenApp):
            _CONTEXTS = {
                'dpset': dpset.DPSet,
            }

            def __init__(self, *args, **kwargs):
                super(MyApp, self).__init__(*args, **kwargs)
                # Stores DPSet instance to call its API in this app
                self.dpset = kwargs['dpset']

            def _my_handler(self):
                # Get the datapath object which has the given dpid
                dpid = 1
                dp = self.dpset.get(dpid)
                if dp is None:
                    self.logger.info('No such datapath: dpid=%d', dpid)
    """

    def __init__(self, *args, **kwargs):
        super(DPSet, self).__init__(*args, **kwargs)
        self.name = 'dpset'

        self.dps = {}   # datapath_id => class Datapath
        self.port_state = {}  # datapath_id => ports

    def _register(self, dp):
        LOG.debug('DPSET: register datapath %s', dp)
        assert dp.id is not None

        # while dpid should be unique, we need to handle duplicates here
        # because it's entirely possible for a switch to reconnect us
        # before we notice the drop of the previous connection.
        # in that case,
        # - forget the older connection as it likely will disappear soon
        # - do not send EventDP leave/enter events
        # - keep the PortState for the dpid
        send_dp_reconnected = False
        if dp.id in self.dps:
            self.logger.warning('DPSET: Multiple connections from %s',
                                dpid_to_str(dp.id))
            self.logger.debug('DPSET: Forgetting datapath %s', self.dps[dp.id])
            (self.dps[dp.id]).close()
            self.logger.debug('DPSET: New datapath %s', dp)
            send_dp_reconnected = True
        self.dps[dp.id] = dp
        if dp.id not in self.port_state:
            self.port_state[dp.id] = PortState()
            ev = EventDP(dp, True)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                for port in dp.ports.values():
                    self._port_added(dp, port)
                    ev.ports.append(port)
            self.send_event_to_observers(ev)
        if send_dp_reconnected:
            ev = EventDPReconnected(dp)
            ev.ports = self.port_state.get(dp.id, {}).values()
            self.send_event_to_observers(ev)

    def _unregister(self, dp):
        # see the comment in _register().
        if dp not in self.dps.values():
            return
        LOG.debug('DPSET: unregister datapath %s', dp)
        assert self.dps[dp.id] == dp

        # Now datapath is already dead, so port status change event doesn't
        # interfere us.
        ev = EventDP(dp, False)
        for port in list(self.port_state.get(dp.id, {}).values()):
            self._port_deleted(dp, port)
            ev.ports.append(port)

        self.send_event_to_observers(ev)

        del self.dps[dp.id]
        del self.port_state[dp.id]

    def get(self, dp_id):
        """
        This method returns the os_ken.controller.controller.Datapath
        instance for the given Datapath ID.
        """
        return self.dps.get(dp_id)

    def get_all(self):
        """
        This method returns a list of tuples which represents
        instances for switches connected to this controller.
        The tuple consists of a Datapath ID and an instance of
        os_ken.controller.controller.Datapath.

        A return value looks like the following::

            [ (dpid_A, Datapath_A), (dpid_B, Datapath_B), ... ]
        """
        return list(self.dps.items())

    def _port_added(self, datapath, port):
        self.port_state[datapath.id].add(port.port_no, port)

    def _port_deleted(self, datapath, port):
        self.port_state[datapath.id].remove(port.port_no)

    @set_ev_cls(ofp_event.EventOFPStateChange,
                [handler.MAIN_DISPATCHER, handler.DEAD_DISPATCHER])
    def dispatcher_change(self, ev):
        datapath = ev.datapath
        assert datapath is not None
        if ev.state == handler.MAIN_DISPATCHER:
            self._register(datapath)
        elif ev.state == handler.DEAD_DISPATCHER:
            self._unregister(datapath)

    @set_ev_cls(ofp_event.EventOFPSwitchFeatures, handler.CONFIG_DISPATCHER)
    def switch_features_handler(self, ev):
        msg = ev.msg
        datapath = msg.datapath
        # ofp_handler.py does the following so we could remove...
        if datapath.ofproto.OFP_VERSION < 0x04:
            datapath.ports = msg.ports

    @set_ev_cls(ofp_event.EventOFPPortStatus, handler.MAIN_DISPATCHER)
    def port_status_handler(self, ev):
        msg = ev.msg
        reason = msg.reason
        datapath = msg.datapath
        port = msg.desc
        ofproto = datapath.ofproto

        if reason == ofproto.OFPPR_ADD:
            LOG.debug('DPSET: A port was added.' +
                      '(datapath id = %s, port number = %s)',
                      dpid_to_str(datapath.id), port.port_no)
            self._port_added(datapath, port)
            self.send_event_to_observers(EventPortAdd(datapath, port))
        elif reason == ofproto.OFPPR_DELETE:
            LOG.debug('DPSET: A port was deleted.' +
                      '(datapath id = %s, port number = %s)',
                      dpid_to_str(datapath.id), port.port_no)
            self._port_deleted(datapath, port)
            self.send_event_to_observers(EventPortDelete(datapath, port))
        else:
            assert reason == ofproto.OFPPR_MODIFY
            LOG.debug('DPSET: A port was modified.' +
                      '(datapath id = %s, port number = %s)',
                      dpid_to_str(datapath.id), port.port_no)
            self.port_state[datapath.id].modify(port.port_no, port)
            self.send_event_to_observers(EventPortModify(datapath, port))

    def get_port(self, dpid, port_no):
        """
        This method returns the os_ken.controller.dpset.PortState
        instance for the given Datapath ID and the port number.
        Raises os_ken_exc.PortNotFound if no such a datapath connected to
        this controller or no such a port exists.
        """
        try:
            return self.port_state[dpid][port_no]
        except KeyError:
            raise os_ken_exc.PortNotFound(dpid=dpid, port=port_no,
                                       network_id=None)

    def get_ports(self, dpid):
        """
        This method returns a list of os_ken.controller.dpset.PortState
        instances for the given Datapath ID.
        Raises KeyError if no such a datapath connected to this controller.
        """
        return list(self.port_state[dpid].values())


handler.register_service('os_ken.controller.dpset')



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Module defines protocol based classes and utils.
"""

from abc import ABCMeta
from abc import abstractmethod
import six


@six.add_metaclass(ABCMeta)
class Protocol(object):
    """Interface for various protocols.

    Protocol usually encloses a transport/connection/socket to
    peer/client/server and encodes and decodes communication/messages. Protocol
    can also maintain any related state machine, protocol message encoding or
    decoding utilities. This interface identifies minimum methods to support to
    facilitate or provide hooks to sub-classes to override behavior as
    appropriate.
    """

    @abstractmethod
    def data_received(self, data):
        """Handler for date received over connection/transport.

        Here *data* is in raw bytes. This *data* should further be converted to
        protocol specific messages and as appropriate transition to new state
        machine state or send appropriate response.
        """
        pass

    @abstractmethod
    def connection_made(self):
        """Called when connection has been established according to protocol.

        This is the right place to do some initialization or sending initial
        hello messages.
        """
        pass

    @abstractmethod
    def connection_lost(self, reason):
        """Handler called when connection to peer/remote according to protocol
        has been lost.

        Here we can do any clean-up related to connection/transport/timers/etc.
        """
        pass


@six.add_metaclass(ABCMeta)
class Factory(object):
    """This is a factory which produces protocols.

    Can also act as context for protocols.
    """

    # Put a subclass of Protocol here:
    protocol = None

    @abstractmethod
    def build_protocol(self, socket):
        """Create an instance of a subclass of Protocol.

        Override this method to alter how Protocol instances get created.
        """
        raise NotImplementedError()

    @abstractmethod
    def start_protocol(self, socket):
        """Launch protocol instance to handle input on an incoming connection.
        """
        raise NotImplementedError()



# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Stratosphere Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# vim: tabstop=4 shiftwidth=4 softtabstop=4

import json
import os
import ssl
import sys
import warnings
import logging
import random
import unittest
from unittest import mock

from nose.tools import eq_, raises

from os_ken.base import app_manager  # To suppress cyclic import
from os_ken.controller import controller
from os_ken.controller import handler
from os_ken.lib import hub
from os_ken.ofproto import ofproto_v1_3_parser
from os_ken.ofproto import ofproto_v1_2_parser
from os_ken.ofproto import ofproto_v1_0_parser
hub.patch()


LOG = logging.getLogger('test_controller')


class TestUtils(unittest.TestCase):
    """
    Test cases for utilities defined in controller module.
    """

    def test_split_addr_with_ipv4(self):
        addr, port = controller._split_addr('127.0.0.1:6653')
        eq_('127.0.0.1', addr)
        eq_(6653, port)

    def test_split_addr_with_ipv6(self):
        addr, port = controller._split_addr('[::1]:6653')
        eq_('::1', addr)
        eq_(6653, port)

    @raises(ValueError)
    def test_split_addr_with_invalid_addr(self):
        controller._split_addr('127.0.0.1')

    @raises(ValueError)
    def test_split_addr_with_invalid_ipv4_addr(self):
        controller._split_addr('xxx.xxx.xxx.xxx:6653')

    @raises(ValueError)
    def test_split_addr_with_invalid_ipv6_addr(self):
        controller._split_addr('[::xxxx]:6653')

    @raises(ValueError)
    def test_split_addr_with_non_bracketed_ipv6_addr(self):
        controller._split_addr('::1:6653')


class Test_Datapath(unittest.TestCase):
    """
    Test cases for controller.Datapath
    """

    def _test_ports_accessibility(self, ofproto_parser, msgs_len):
        with mock.patch('os_ken.controller.controller.Datapath.set_state'):

            # Ignore warnings
            with warnings.catch_warnings(record=True) as msgs:
                warnings.simplefilter('always')

                # Test target
                sock_mock = mock.Mock()
                addr_mock = mock.Mock()
                dp = controller.Datapath(sock_mock, addr_mock)
                dp.ofproto_parser = ofproto_parser

                # Create
                dp.ports = {}

                # Update
                port_mock = mock.Mock()
                dp.ports[0] = port_mock

                # Read & Delete
                del dp.ports[0]

                self.assertEqual(len(msgs), msgs_len)
                for msg in msgs:
                    self.assertTrue(issubclass(msg.category, UserWarning))

    def test_ports_accessibility_v13(self):
        self._test_ports_accessibility(ofproto_v1_3_parser, 0)

    def test_ports_accessibility_v12(self):
        self._test_ports_accessibility(ofproto_v1_2_parser, 0)

    def test_ports_accessibility_v10(self):
        self._test_ports_accessibility(ofproto_v1_0_parser, 0)

    @mock.patch("os_ken.base.app_manager", spec=app_manager)
    def test_recv_loop(self, app_manager_mock):
        # Prepare test data
        test_messages = [
            "4-6-ofp_features_reply.packet",
            "4-14-ofp_echo_reply.packet",
            "4-14-ofp_echo_reply.packet",
            "4-4-ofp_packet_in.packet",
            "4-14-ofp_echo_reply.packet",
            "4-14-ofp_echo_reply.packet",
        ]
        this_dir = os.path.dirname(sys.modules[__name__].__file__)
        packet_data_dir = os.path.join(this_dir, '../../packet_data/of13')
        json_dir = os.path.join(this_dir, '../ofproto/json/of13')

        packet_buf = bytearray()
        expected_json = list()
        for msg in test_messages:
            # Construct the received packet buffer as one packet data in order
            # to test the case of the OpenFlow messages composed in one packet.
            packet_data_file = os.path.join(packet_data_dir, msg)
            packet_buf += open(packet_data_file, 'rb').read()
            json_data_file = os.path.join(json_dir, msg + '.json')
            expected_json.append(json.load(open(json_data_file)))

        # Prepare mock for socket
        class SocketMock(mock.MagicMock):
            buf = bytearray()
            random = None

            def recv(self, bufsize):
                size = self.random.randint(1, bufsize)
                out = self.buf[:size]
                self.buf = self.buf[size:]
                return out

        # Prepare mock
        ofp_brick_mock = mock.MagicMock(spec=app_manager.OSKenApp)
        app_manager_mock.lookup_service_brick.return_value = ofp_brick_mock
        sock_mock = SocketMock()
        sock_mock.buf = packet_buf
        sock_mock.random = random.Random('OSKen SDN Framework')
        addr_mock = mock.MagicMock()

        # Prepare test target
        dp = controller.Datapath(sock_mock, addr_mock)
        dp.set_state(handler.MAIN_DISPATCHER)
        ofp_brick_mock.reset_mock()

        # Test
        dp._recv_loop()

        # Assert calls
        output_json = list()
        for call in ofp_brick_mock.send_event_to_observers.call_args_list:
            args, kwargs = call
            ev, state = args
            if not hasattr(ev, 'msg'):
                continue
            output_json.append(ev.msg.to_jsondict())
            self.assertEqual(state, handler.MAIN_DISPATCHER)
            self.assertEqual(kwargs, {})
        self.assertEqual(expected_json, output_json)


class TestOpenFlowController(unittest.TestCase):
    """
    Test cases for OpenFlowController
    """
    @mock.patch("os_ken.controller.controller.CONF")
    def _test_ssl(self, this_dir, port, conf_mock):
        conf_mock.ofp_ssl_listen_port = port
        conf_mock.ofp_listen_host = "127.0.0.1"
        conf_mock.ca_certs = None
        conf_mock.ciphers = None
        conf_mock.ctl_cert = os.path.join(this_dir, 'cert.crt')
        conf_mock.ctl_privkey = os.path.join(this_dir, 'cert.key')
        c = controller.OpenFlowController()
        c()

    def test_ssl(self):
        """Tests SSL server functionality."""
        # TODO: TLS version enforcement is necessary to avoid
        # vulnerable versions. Currently, this only tests TLS
        # connectivity.
        this_dir = os.path.dirname(sys.modules[__name__].__file__)
        saved_exception = None
        try:
            ssl_version = ssl.PROTOCOL_TLS
        except AttributeError:
            # For compatibility with older pythons.
            ssl_version = ssl.PROTOCOL_TLSv1
        for i in range(3):
            try:
                # Try a few times as this can fail with EADDRINUSE
                port = random.randint(5000, 10000)
                server = hub.spawn(self._test_ssl, this_dir, port)
                hub.sleep(1)
                client = hub.StreamClient(("127.0.0.1", port),
                                          timeout=5,
                                          ssl_version=ssl_version)
                if client.connect() is not None:
                    break
            except Exception as e:
                saved_exception = e
                continue
            finally:
                try:
                    hub.kill(server)
                except Exception:
                    pass
        else:
            self.fail("Failed to connect: " + str(saved_exception))



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import logging
import inspect
import six
import struct

from nose.tools import *
from os_ken.lib import addrconv
from os_ken.lib.packet import cfm

LOG = logging.getLogger(__name__)


class Test_cfm(unittest.TestCase):

    def setUp(self):
        self.message = cfm.cc_message()
        self.ins = cfm.cfm(self.message)
        data = bytearray()
        prev = None
        self.buf = self.ins.serialize(data, prev)

    def setUp_cc_message(self):
        self.cc_message_md_lv = 1
        self.cc_message_version = 1
        self.cc_message_rdi = 1
        self.cc_message_interval = 1
        self.cc_message_seq_num = 123
        self.cc_message_mep_id = 4
        self.cc_message_md_name_format = 4
        self.cc_message_md_name_length = 0
        self.cc_message_md_name = b"hoge"
        self.cc_message_short_ma_name_format = 2
        self.cc_message_short_ma_name_length = 0
        self.cc_message_short_ma_name = b"pakeratta"
        self.cc_message_md_name_txfcf = 11
        self.cc_message_md_name_rxfcb = 22
        self.cc_message_md_name_txfcb = 33
        self.cc_message_tlvs = [
            cfm.sender_id_tlv(),
            cfm.port_status_tlv(),
            cfm.data_tlv(),
            cfm.interface_status_tlv(),
            cfm.reply_ingress_tlv(),
            cfm.reply_egress_tlv(),
            cfm.ltm_egress_identifier_tlv(),
            cfm.ltr_egress_identifier_tlv(),
            cfm.organization_specific_tlv(),
        ]
        self.message = cfm.cc_message(
            self.cc_message_md_lv,
            self.cc_message_version,
            self.cc_message_rdi,
            self.cc_message_interval,
            self.cc_message_seq_num,
            self.cc_message_mep_id,
            self.cc_message_md_name_format,
            self.cc_message_md_name_length,
            self.cc_message_md_name,
            self.cc_message_short_ma_name_format,
            self.cc_message_short_ma_name_length,
            self.cc_message_short_ma_name,
            self.cc_message_tlvs
        )
        self.ins = cfm.cfm(self.message)
        data = bytearray()
        prev = None
        self.buf = self.ins.serialize(data, prev)

    def setUp_loopback_message(self):
        self.loopback_message_md_lv = 1
        self.loopback_message_version = 1
        self.loopback_message_transaction_id = 12345
        self.loopback_message_tlvs = [
            cfm.sender_id_tlv(),
            cfm.port_status_tlv(),
            cfm.data_tlv(),
            cfm.interface_status_tlv(),
            cfm.reply_ingress_tlv(),
            cfm.reply_egress_tlv(),
            cfm.ltm_egress_identifier_tlv(),
            cfm.ltr_egress_identifier_tlv(),
            cfm.organization_specific_tlv(),
        ]
        self.message = cfm.loopback_message(
            self.loopback_message_md_lv,
            self.loopback_message_version,
            self.loopback_message_transaction_id,
            self.loopback_message_tlvs)
        self.ins = cfm.cfm(self.message)
        data = bytearray()
        prev = None
        self.buf = self.ins.serialize(data, prev)

    def setUp_loopback_reply(self):
        self.loopback_reply_md_lv = 1
        self.loopback_reply_version = 1
        self.loopback_reply_transaction_id = 12345
        self.loopback_reply_tlvs = [
            cfm.sender_id_tlv(),
            cfm.port_status_tlv(),
            cfm.data_tlv(),
            cfm.interface_status_tlv(),
            cfm.reply_ingress_tlv(),
            cfm.reply_egress_tlv(),
            cfm.ltm_egress_identifier_tlv(),
            cfm.ltr_egress_identifier_tlv(),
            cfm.organization_specific_tlv(),
        ]
        self.message = cfm.loopback_reply(
            self.loopback_reply_md_lv,
            self.loopback_reply_version,
            self.loopback_reply_transaction_id,
            self.loopback_reply_tlvs)
        self.ins = cfm.cfm(self.message)
        data = bytearray()
        prev = None
        self.buf = self.ins.serialize(data, prev)

    def setUp_link_trace_message(self):
        self.link_trace_message_md_lv = 1
        self.link_trace_message_version = 1
        self.link_trace_message_use_fdb_only = 1
        self.link_trace_message_transaction_id = 12345
        self.link_trace_message_ttl = 123
        self.link_trace_message_ltm_orig_addr = '11:22:33:44:55:66'
        self.link_trace_message_ltm_targ_addr = '77:88:99:aa:cc:dd'
        self.link_trace_message_tlvs = [
            cfm.sender_id_tlv(),
            cfm.port_status_tlv(),
            cfm.data_tlv(),
            cfm.interface_status_tlv(),
            cfm.reply_ingress_tlv(),
            cfm.reply_egress_tlv(),
            cfm.ltm_egress_identifier_tlv(),
            cfm.ltr_egress_identifier_tlv(),
            cfm.organization_specific_tlv(),
        ]
        self.message = cfm.link_trace_message(
            self.link_trace_message_md_lv,
            self.link_trace_message_version,
            self.link_trace_message_use_fdb_only,
            self.link_trace_message_transaction_id,
            self.link_trace_message_ttl,
            self.link_trace_message_ltm_orig_addr,
            self.link_trace_message_ltm_targ_addr,
            self.link_trace_message_tlvs)
        self.ins = cfm.cfm(self.message)
        data = bytearray()
        prev = None
        self.buf = self.ins.serialize(data, prev)

    def setUp_link_trace_reply(self):
        self.link_trace_reply_md_lv = 1
        self.link_trace_reply_version = 1
        self.link_trace_reply_use_fdb_only = 1
        self.link_trace_reply_fwd_yes = 0
        self.link_trace_reply_terminal_mep = 1
        self.link_trace_reply_transaction_id = 5432
        self.link_trace_reply_ttl = 123
        self.link_trace_reply_relay_action = 3
        self.link_trace_reply_tlvs = [
            cfm.sender_id_tlv(),
            cfm.port_status_tlv(),
            cfm.data_tlv(),
            cfm.interface_status_tlv(),
            cfm.reply_ingress_tlv(),
            cfm.reply_egress_tlv(),
            cfm.ltm_egress_identifier_tlv(),
            cfm.ltr_egress_identifier_tlv(),
            cfm.organization_specific_tlv(),
        ]
        self.message = cfm.link_trace_reply(
            self.link_trace_reply_md_lv,
            self.link_trace_reply_version,
            self.link_trace_reply_use_fdb_only,
            self.link_trace_reply_fwd_yes,
            self.link_trace_reply_terminal_mep,
            self.link_trace_reply_transaction_id,
            self.link_trace_reply_ttl,
            self.link_trace_reply_relay_action,
            self.link_trace_reply_tlvs)
        self.ins = cfm.cfm(self.message)
        data = bytearray()
        prev = None
        self.buf = self.ins.serialize(data, prev)

    def tearDown(self):
        pass

    def test_init(self):
        eq_(str(self.message), str(self.ins.op))

    def test_init_cc_message(self):
        self.setUp_cc_message()
        self.test_init()

    def test_init_loopback_message(self):
        self.setUp_loopback_message()
        self.test_init()

    def test_init_loopback_reply(self):
        self.setUp_loopback_reply()
        self.test_init()

    def test_init_link_trace_message(self):
        self.setUp_link_trace_message()
        self.test_init()

    def test_init_link_trace_reply(self):
        self.setUp_link_trace_reply()
        self.test_init()

    def test_parser(self):
        _res = self.ins.parser(six.binary_type(self.buf))

        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(str(self.message), str(res.op))

    def test_parser_with_cc_message(self):
        self.setUp_cc_message()
        self.test_parser()

    def test_parser_with_loopback_message(self):
        self.setUp_loopback_message()
        self.test_parser()

    def test_parser_with_loopback_reply(self):
        self.setUp_loopback_reply()
        self.test_parser()

    def test_parser_with_link_trace_message(self):
        self.setUp_link_trace_message()
        self.test_parser()

    def test_parser_with_link_trace_reply(self):
        self.setUp_link_trace_reply()
        self.test_parser()

    def test_serialize(self):
        pass

    def test_serialize_with_cc_message(self):
        self.setUp_cc_message()
        self.test_serialize()
        data = bytearray()
        prev = None
        buf = self.ins.serialize(data, prev)
        cc_message = cfm.cc_message.parser(six.binary_type(buf))
        eq_(repr(self.message), repr(cc_message))

    def test_serialize_with_loopback_message(self):
        self.setUp_loopback_message()
        self.test_serialize()
        data = bytearray()
        prev = None
        buf = self.ins.serialize(data, prev)
        loopback_message = cfm.loopback_message.parser(six.binary_type(buf))
        eq_(repr(self.message), repr(loopback_message))

    def test_serialize_with_loopback_reply(self):
        self.setUp_loopback_reply()
        self.test_serialize()
        data = bytearray()
        prev = None
        buf = self.ins.serialize(data, prev)
        loopback_reply = cfm.loopback_reply.parser(six.binary_type(buf))
        eq_(repr(self.message), repr(loopback_reply))

    def test_serialize_with_link_trace_message(self):
        self.setUp_link_trace_message()
        self.test_serialize()
        data = bytearray()
        prev = None
        buf = self.ins.serialize(data, prev)
        link_trace_message = cfm.link_trace_message.parser(six.binary_type(buf))
        eq_(repr(self.message), repr(link_trace_message))

    def test_serialize_with_link_trace_reply(self):
        self.setUp_link_trace_reply()
        self.test_serialize()
        data = bytearray()
        prev = None
        buf = self.ins.serialize(data, prev)
        link_trace_reply = cfm.link_trace_reply.parser(six.binary_type(buf))
        eq_(repr(self.message), repr(link_trace_reply))

    def test_to_string(self):
        cfm_values = {'op': self.message}
        _cfm_str = ','.join(['%s=%s' % (k, cfm_values[k])
                             for k, v in inspect.getmembers(self.ins)
                             if k in cfm_values])
        cfm_str = '%s(%s)' % (cfm.cfm.__name__, _cfm_str)
        eq_(str(self.ins), cfm_str)
        eq_(repr(self.ins), cfm_str)

    def test_to_string_cc_message(self):
        self.setUp_cc_message()
        self.test_to_string()

    def test_to_string_loopback_message(self):
        self.setUp_loopback_message()
        self.test_to_string()

    def test_to_string_loopback_reply(self):
        self.setUp_loopback_reply()
        self.test_to_string()

    def test_to_string_link_trace_message(self):
        self.setUp_link_trace_message()
        self.test_to_string()

    def test_to_string_link_trace_reply(self):
        self.setUp_link_trace_reply()
        self.test_to_string()

    def test_len(self):
        pass

    def test_len_cc_message(self):
        self.setUp_cc_message()
        eq_(len(self.ins), 0 + len(self.message))

    def test_len_loopback_message(self):
        self.setUp_loopback_message()
        eq_(len(self.ins), 0 + len(self.message))

    def test_len_loopback_reply(self):
        self.setUp_loopback_reply()
        eq_(len(self.ins), 0 + len(self.message))

    def test_len_link_trace_message(self):
        self.setUp_link_trace_message()
        eq_(len(self.ins), 0 + len(self.message))

    def test_len_link_trace_reply(self):
        self.setUp_link_trace_reply()
        eq_(len(self.ins), 0 + len(self.message))

    def test_default_args(self):
        pass

    def test_json(self):
        jsondict = self.ins.to_jsondict()
        ins = cfm.cfm.from_jsondict(jsondict['cfm'])
        eq_(str(self.ins), str(ins))

    def test_json_with_cc_message(self):
        self.setUp_cc_message()
        self.test_json()

    def test_json_with_loopback_message(self):
        self.setUp_loopback_message()
        self.test_json()

    def test_json_with_loopback_reply(self):
        self.setUp_loopback_reply()
        self.test_json()

    def test_json_with_link_trace_message(self):
        self.setUp_link_trace_message()
        self.test_json()

    def test_json_with_link_trace_reply(self):
        self.setUp_link_trace_reply()
        self.test_json()


class Test_cc_message(unittest.TestCase):

    def setUp(self):
        self.md_lv = 1
        self.version = 1
        self.opcode = cfm.CFM_CC_MESSAGE
        self.rdi = 1
        self.interval = 5
        self.first_tlv_offset = cfm.cc_message._TLV_OFFSET
        self.seq_num = 2
        self.mep_id = 2
        self.md_name_format = cfm.cc_message._MD_FMT_CHARACTER_STRING
        self.md_name_length = 3
        self.md_name = b"foo"
        self.short_ma_name_format = 2
        self.short_ma_name_length = 8
        self.short_ma_name = b"hogehoge"
        self.tlvs = [
        ]
        self.end_tlv = 0
        self.ins = cfm.cc_message(
            self.md_lv,
            self.version,
            self.rdi,
            self.interval,
            self.seq_num,
            self.mep_id,
            self.md_name_format,
            self.md_name_length,
            self.md_name,
            self.short_ma_name_format,
            self.short_ma_name_length,
            self.short_ma_name,
            self.tlvs
        )

        self.form = '!4BIH2B3s2B8s33x12x4xB'
        self.buf = struct.pack(
            self.form,
            (self.md_lv << 5) | self.version,
            self.opcode,
            (self.rdi << 7) | self.interval,
            self.first_tlv_offset,
            self.seq_num,
            self.mep_id,
            self.md_name_format,
            self.md_name_length,
            self.md_name,
            self.short_ma_name_format,
            self.short_ma_name_length,
            self.short_ma_name,
            self.end_tlv
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.md_lv, self.ins.md_lv)
        eq_(self.version, self.ins.version)
        eq_(self.rdi, self.ins.rdi)
        eq_(self.interval, self.ins.interval)
        eq_(self.seq_num, self.ins.seq_num)
        eq_(self.mep_id, self.ins.mep_id)
        eq_(self.md_name_format, self.ins.md_name_format)
        eq_(self.md_name_length, self.ins.md_name_length)
        eq_(self.md_name, self.ins.md_name)
        eq_(self.short_ma_name_format, self.ins.short_ma_name_format)
        eq_(self.short_ma_name_length, self.ins.short_ma_name_length)
        eq_(self.short_ma_name, self.ins.short_ma_name)
        eq_(self.tlvs, self.ins.tlvs)

    def test_parser(self):
        _res = cfm.cc_message.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.md_lv, res.md_lv)
        eq_(self.version, res.version)
        eq_(self.rdi, res.rdi)
        eq_(self.interval, res.interval)
        eq_(self.seq_num, res.seq_num)
        eq_(self.mep_id, res.mep_id)
        eq_(self.md_name_format, res.md_name_format)
        eq_(self.md_name_length, res.md_name_length)
        eq_(self.md_name, res.md_name)
        eq_(self.short_ma_name_format, res.short_ma_name_format)
        eq_(self.short_ma_name_length, res.short_ma_name_length)
        eq_(self.short_ma_name, res.short_ma_name)
        eq_(self.tlvs, res.tlvs)

    def test_parser_with_no_maintenance_domain_name_present(self):
        form = '!4BIH3B8s37x12x4xB'
        buf = struct.pack(
            form,
            (self.md_lv << 5) | self.version,
            self.opcode,
            (self.rdi << 7) | self.interval,
            self.first_tlv_offset,
            self.seq_num,
            self.mep_id,
            cfm.cc_message._MD_FMT_NO_MD_NAME_PRESENT,
            self.short_ma_name_format,
            self.short_ma_name_length,
            self.short_ma_name,
            self.end_tlv
        )
        _res = cfm.cc_message.parser(buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.md_lv, res.md_lv)
        eq_(self.version, res.version)
        eq_(self.rdi, res.rdi)
        eq_(self.interval, res.interval)
        eq_(self.seq_num, res.seq_num)
        eq_(self.mep_id, res.mep_id)
        eq_(cfm.cc_message._MD_FMT_NO_MD_NAME_PRESENT, res.md_name_format)
        eq_(self.short_ma_name_format, res.short_ma_name_format)
        eq_(self.short_ma_name_length, res.short_ma_name_length)
        eq_(self.short_ma_name, res.short_ma_name)
        eq_(self.tlvs, res.tlvs)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self.md_lv, res[0] >> 5)
        eq_(self.version, res[0] & 0x1f)
        eq_(self.opcode, res[1])
        eq_(self.rdi, res[2] >> 7)
        eq_(self.interval, res[2] & 0x07)
        eq_(self.first_tlv_offset, res[3])
        eq_(self.seq_num, res[4])
        eq_(self.mep_id, res[5])
        eq_(self.md_name_format, res[6])
        eq_(self.md_name_length, res[7])
        eq_(self.md_name, res[8])
        eq_(self.short_ma_name_format, res[9])
        eq_(self.short_ma_name_length, res[10])
        eq_(self.short_ma_name, res[11])
        eq_(self.end_tlv, res[12])

    def test_serialize_with_md_name_length_zero(self):
        ins = cfm.cc_message(
            self.md_lv,
            self.version,
            self.rdi,
            self.interval,
            self.seq_num,
            self.mep_id,
            self.md_name_format,
            0,
            self.md_name,
            self.short_ma_name_format,
            0,
            self.short_ma_name,
            self.tlvs
        )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self.md_lv, res[0] >> 5)
        eq_(self.version, res[0] & 0x1f)
        eq_(self.opcode, res[1])
        eq_(self.rdi, res[2] >> 7)
        eq_(self.interval, res[2] & 0x07)
        eq_(self.first_tlv_offset, res[3])
        eq_(self.seq_num, res[4])
        eq_(self.mep_id, res[5])
        eq_(self.md_name_format, res[6])
        eq_(self.md_name_length, res[7])
        eq_(self.md_name, res[8])
        eq_(self.short_ma_name_format, res[9])
        eq_(self.short_ma_name_length, res[10])
        eq_(self.short_ma_name, res[11])
        eq_(self.end_tlv, res[12])

    def test_serialize_with_no_maintenance_domain_name_present(self):
        form = '!4BIH3B8s37x12x4xB'
        ins = cfm.cc_message(
            self.md_lv,
            self.version,
            self.rdi,
            self.interval,
            self.seq_num,
            self.mep_id,
            cfm.cc_message._MD_FMT_NO_MD_NAME_PRESENT,
            0,
            self.md_name,
            self.short_ma_name_format,
            0,
            self.short_ma_name,
            self.tlvs
        )
        buf = ins.serialize()
        res = struct.unpack_from(form, six.binary_type(buf))
        eq_(self.md_lv, res[0] >> 5)
        eq_(self.version, res[0] & 0x1f)
        eq_(self.opcode, res[1])
        eq_(self.rdi, res[2] >> 7)
        eq_(self.interval, res[2] & 0x07)
        eq_(self.first_tlv_offset, res[3])
        eq_(self.seq_num, res[4])
        eq_(self.mep_id, res[5])
        eq_(cfm.cc_message._MD_FMT_NO_MD_NAME_PRESENT, res[6])
        eq_(self.short_ma_name_format, res[7])
        eq_(self.short_ma_name_length, res[8])
        eq_(self.short_ma_name, res[9])
        eq_(self.end_tlv, res[10])

    def test_len(self):
        # 75 octet (If tlv does not exist)
        eq_(75, len(self.ins))

    def test_default_args(self):
        ins = cfm.cc_message()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.cc_message._PACK_STR, six.binary_type(buf))
        eq_(res[0] >> 5, 0)
        eq_(res[0] & 0x1f, 0)
        eq_(res[1], 1)
        eq_(res[2] >> 7, 0)
        eq_(res[2] & 0x07, 4)
        eq_(res[3], 70)
        eq_(res[4], 0)
        eq_(res[5], 1)
        eq_(res[6], 4)


class Test_loopback_message(unittest.TestCase):

    def setUp(self):
        self.md_lv = 1
        self.version = 1
        self.opcode = cfm.CFM_LOOPBACK_MESSAGE
        self.flags = 0
        self.first_tlv_offset = cfm.loopback_message._TLV_OFFSET
        self.transaction_id = 12345
        self.tlvs = [
        ]

        self.end_tlv = 0
        self.ins = cfm.loopback_message(
            self.md_lv,
            self.version,
            self.transaction_id,
            self.tlvs,
        )
        self.form = '!4BIB'
        self.buf = struct.pack(
            self.form,
            (self.md_lv << 5) | self.version,
            self.opcode,
            self.flags,
            self.first_tlv_offset,
            self.transaction_id,
            self.end_tlv
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.md_lv, self.ins.md_lv)
        eq_(self.version, self.ins.version)
        eq_(self.transaction_id, self.ins.transaction_id)
        eq_(self.tlvs, self.ins.tlvs)

    def test_parser(self):
        _res = cfm.loopback_message.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.md_lv, res.md_lv)
        eq_(self.version, res.version)
        eq_(self.transaction_id, res.transaction_id)
        eq_(self.tlvs, res.tlvs)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self.md_lv, res[0] >> 5)
        eq_(self.version, res[0] & 0x1f)
        eq_(self.opcode, res[1])
        eq_(self.flags, res[2])
        eq_(self.first_tlv_offset, res[3])
        eq_(self.transaction_id, res[4])
        eq_(self.end_tlv, res[5])

    def test_len(self):
        # 9 octet (If tlv does not exist)
        eq_(9, len(self.ins))

    def test_default_args(self):
        ins = cfm.loopback_message()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.loopback_message._PACK_STR,
                                 six.binary_type(buf))
        eq_(res[0] >> 5, 0)
        eq_(res[0] & 0x1f, 0)
        eq_(res[1], 3)
        eq_(res[2], 0)
        eq_(res[3], 4)
        eq_(res[4], 0)


class Test_loopback_reply(unittest.TestCase):

    def setUp(self):
        self.md_lv = 1
        self.version = 1
        self.opcode = cfm.CFM_LOOPBACK_REPLY
        self.flags = 0
        self.first_tlv_offset = cfm.loopback_reply._TLV_OFFSET
        self.transaction_id = 12345
        self.tlvs = [
        ]
        self.end_tlv = 0
        self.ins = cfm.loopback_reply(
            self.md_lv,
            self.version,
            self.transaction_id,
            self.tlvs,
        )
        self.form = '!4BIB'
        self.buf = struct.pack(
            self.form,
            (self.md_lv << 5) | self.version,
            self.opcode,
            self.flags,
            self.first_tlv_offset,
            self.transaction_id,
            self.end_tlv
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.md_lv, self.ins.md_lv)
        eq_(self.version, self.ins.version)
        eq_(self.transaction_id, self.ins.transaction_id)
        eq_(self.tlvs, self.ins.tlvs)

    def test_parser(self):
        _res = cfm.loopback_reply.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.md_lv, res.md_lv)
        eq_(self.version, res.version)
        eq_(self.transaction_id, res.transaction_id)
        eq_(self.tlvs, res.tlvs)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self.md_lv, res[0] >> 5)
        eq_(self.version, res[0] & 0x1f)
        eq_(self.opcode, res[1])
        eq_(self.flags, res[2])
        eq_(self.first_tlv_offset, res[3])
        eq_(self.transaction_id, res[4])
        eq_(self.end_tlv, res[5])

    def test_len(self):
        # 9 octet (If tlv does not exist)
        eq_(9, len(self.ins))

    def test_default_args(self):
        ins = cfm.loopback_reply()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.loopback_reply._PACK_STR, six.binary_type(buf))
        eq_(res[0] >> 5, 0)
        eq_(res[0] & 0x1f, 0)
        eq_(res[1], 2)
        eq_(res[2], 0)
        eq_(res[3], 4)
        eq_(res[4], 0)


class Test_link_trace_message(unittest.TestCase):

    def setUp(self):
        self.md_lv = 1
        self.version = 1
        self.opcode = cfm.CFM_LINK_TRACE_MESSAGE
        self.use_fdb_only = 1
        self.first_tlv_offset = cfm.link_trace_message._TLV_OFFSET
        self.transaction_id = 12345
        self.ttl = 55
        self.ltm_orig_addr = "00:11:22:44:55:66"
        self.ltm_targ_addr = "ab:cd:ef:23:12:65"
        self.tlvs = [
        ]

        self.end_tlv = 0
        self.ins = cfm.link_trace_message(
            self.md_lv,
            self.version,
            self.use_fdb_only,
            self.transaction_id,
            self.ttl,
            self.ltm_orig_addr,
            self.ltm_targ_addr,
            self.tlvs
        )
        self.form = '!4BIB6s6sB'
        self.buf = struct.pack(
            self.form,
            (self.md_lv << 5) | self.version,
            self.opcode,
            self.use_fdb_only << 7,
            self.first_tlv_offset,
            self.transaction_id,
            self.ttl,
            addrconv.mac.text_to_bin(self.ltm_orig_addr),
            addrconv.mac.text_to_bin(self.ltm_targ_addr),
            self.end_tlv
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.md_lv, self.ins.md_lv)
        eq_(self.version, self.ins.version)
        eq_(self.use_fdb_only, self.ins.use_fdb_only)
        eq_(self.transaction_id, self.ins.transaction_id)
        eq_(self.ttl, self.ins.ttl)
        eq_(self.ltm_orig_addr, self.ins.ltm_orig_addr)
        eq_(self.ltm_targ_addr, self.ins.ltm_targ_addr)
        eq_(self.tlvs, self.ins.tlvs)

    def test_parser(self):
        _res = cfm.link_trace_message.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.md_lv, res.md_lv)
        eq_(self.version, res.version)
        eq_(self.use_fdb_only, res.use_fdb_only)
        eq_(self.transaction_id, res.transaction_id)
        eq_(self.ttl, res.ttl)
        eq_(self.ltm_orig_addr, res.ltm_orig_addr)
        eq_(self.ltm_targ_addr, res.ltm_targ_addr)
        eq_(self.tlvs, res.tlvs)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self.md_lv, res[0] >> 5)
        eq_(self.version, res[0] & 0x1f)
        eq_(self.opcode, res[1])
        eq_(self.use_fdb_only, res[2] >> 7)
        eq_(self.first_tlv_offset, res[3])
        eq_(self.transaction_id, res[4])
        eq_(self.ttl, res[5])
        eq_(addrconv.mac.text_to_bin(self.ltm_orig_addr), res[6])
        eq_(addrconv.mac.text_to_bin(self.ltm_targ_addr), res[7])
        eq_(self.end_tlv, res[8])

    def test_len(self):
        # 22 octet (If tlv does not exist)
        eq_(22, len(self.ins))

    def test_default_args(self):
        ins = cfm.link_trace_message()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.link_trace_message._PACK_STR, six.binary_type(buf))
        eq_(res[0] >> 5, 0)
        eq_(res[0] & 0x1f, 0)
        eq_(res[1], 5)
        eq_(res[2] >> 7, 1)
        eq_(res[3], 17)
        eq_(res[4], 0)
        eq_(res[5], 64)
        eq_(res[6], addrconv.mac.text_to_bin('00:00:00:00:00:00'))
        eq_(res[7], addrconv.mac.text_to_bin('00:00:00:00:00:00'))


class Test_link_trace_reply(unittest.TestCase):

    def setUp(self):
        self.md_lv = 1
        self.version = 1
        self.opcode = cfm.CFM_LINK_TRACE_REPLY
        self.use_fdb_only = 1
        self.fwd_yes = 0
        self.terminal_mep = 1
        self.first_tlv_offset = cfm.link_trace_reply._TLV_OFFSET
        self.transaction_id = 12345
        self.ttl = 55
        self.relay_action = 2
        self.ltm_orig_addr = "00:11:22:aa:bb:cc"
        self.ltm_targ_addr = "53:45:24:64:ac:ff"
        self.tlvs = [
        ]
        self.end_tlv = 0
        self.ins = cfm.link_trace_reply(
            self.md_lv,
            self.version,
            self.use_fdb_only,
            self.fwd_yes,
            self.terminal_mep,
            self.transaction_id,
            self.ttl,
            self.relay_action,
            self.tlvs,
        )
        self.form = '!4BIBBB'
        self.buf = struct.pack(
            self.form,
            (self.md_lv << 5) | self.version,
            self.opcode,
            (self.use_fdb_only << 7) | (self.fwd_yes << 6) |
            (self.terminal_mep << 5),
            self.first_tlv_offset,
            self.transaction_id,
            self.ttl,
            self.relay_action,
            self.end_tlv
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.md_lv, self.ins.md_lv)
        eq_(self.version, self.ins.version)
        eq_(self.use_fdb_only, self.ins.use_fdb_only)
        eq_(self.fwd_yes, self.ins.fwd_yes)
        eq_(self.terminal_mep, self.ins.terminal_mep)
        eq_(self.transaction_id, self.ins.transaction_id)
        eq_(self.ttl, self.ins.ttl)
        eq_(self.relay_action, self.ins.relay_action)
        eq_(self.tlvs, self.ins.tlvs)

    def test_parser(self):
        _res = cfm.link_trace_reply.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.md_lv, res.md_lv)
        eq_(self.version, res.version)
        eq_(self.use_fdb_only, self.ins.use_fdb_only)
        eq_(self.fwd_yes, self.ins.fwd_yes)
        eq_(self.terminal_mep, self.ins.terminal_mep)
        eq_(self.transaction_id, res.transaction_id)
        eq_(self.ttl, res.ttl)
        eq_(self.relay_action, res.relay_action)
        eq_(self.tlvs, res.tlvs)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self.md_lv, res[0] >> 5)
        eq_(self.version, res[0] & 0x1f)
        eq_(self.opcode, res[1])
        eq_(self.use_fdb_only, res[2] >> 7 & 0x01)
        eq_(self.fwd_yes, res[2] >> 6 & 0x01)
        eq_(self.terminal_mep, res[2] >> 5 & 0x01)
        eq_(self.first_tlv_offset, res[3])
        eq_(self.transaction_id, res[4])
        eq_(self.ttl, res[5])
        eq_(self.relay_action, res[6])
        eq_(self.end_tlv, res[7])

    def test_len(self):
        # 11 octet (If tlv does not exist)
        eq_(11, len(self.ins))

    def test_default_args(self):
        ins = cfm.link_trace_reply()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.link_trace_reply._PACK_STR, six.binary_type(buf))
        eq_(res[0] >> 5, 0)
        eq_(res[0] & 0x1f, 0)
        eq_(res[1], 4)
        eq_(res[2] >> 7, 1)
        eq_(res[2] >> 6 & 0x01, 0)
        eq_(res[2] >> 5 & 0x01, 1)
        eq_(res[3], 6)
        eq_(res[4], 0)
        eq_(res[5], 64)
        eq_(res[6], 1)


class Test_sender_id_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_SENDER_ID_TLV
        self.length = 10
        self.chassis_id_length = 1
        self.chassis_id_subtype = 3
        self.chassis_id = b"\x0a"
        self.ma_domain_length = 2
        self.ma_domain = b"\x04\x05"
        self.ma_length = 3
        self.ma = b"\x01\x02\x03"
        self.ins = cfm.sender_id_tlv(
            self.length,
            self.chassis_id_length,
            self.chassis_id_subtype,
            self.chassis_id,
            self.ma_domain_length,
            self.ma_domain,
            self.ma_length,
            self.ma,
        )
        self.form = '!BHBB1sB2sB3s'
        self.buf = struct.pack(
            self.form,
            self._type,
            self.length,
            self.chassis_id_length,
            self.chassis_id_subtype,
            self.chassis_id,
            self.ma_domain_length,
            self.ma_domain,
            self.ma_length,
            self.ma
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.chassis_id_length, self.ins.chassis_id_length)
        eq_(self.chassis_id_subtype, self.ins.chassis_id_subtype)
        eq_(self.chassis_id, self.ins.chassis_id)
        eq_(self.ma_domain_length, self.ins.ma_domain_length)
        eq_(self.ma_domain, self.ins.ma_domain)
        eq_(self.ma_length, self.ins.ma_length)
        eq_(self.ma, self.ins.ma)

    def test_parser(self):
        _res = cfm.sender_id_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.chassis_id_length, res.chassis_id_length)
        eq_(self.chassis_id_subtype, res.chassis_id_subtype)
        eq_(self.chassis_id, res.chassis_id)
        eq_(self.ma_domain_length, res.ma_domain_length)
        eq_(self.ma_domain, res.ma_domain)
        eq_(self.ma_length, res.ma_length)
        eq_(self.ma, res.ma)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.chassis_id_length, res[2])
        eq_(self.chassis_id_subtype, res[3])
        eq_(self.chassis_id, res[4])
        eq_(self.ma_domain_length, res[5])
        eq_(self.ma_domain, res[6])
        eq_(self.ma_length, res[7])
        eq_(self.ma, res[8])

    def test_serialize_semi_normal_ptn1(self):
        ins = cfm.sender_id_tlv(
            chassis_id_subtype=self.chassis_id_subtype,
            chassis_id=self.chassis_id,
            ma_domain=self.ma_domain,
        )
        buf = ins.serialize()
        form = '!BHBB1sB2sB'
        res = struct.unpack_from(form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(7, res[1])
        eq_(self.chassis_id_length, res[2])
        eq_(self.chassis_id_subtype, res[3])
        eq_(self.chassis_id, res[4])
        eq_(self.ma_domain_length, res[5])
        eq_(self.ma_domain, res[6])
        eq_(0, res[7])

    def test_serialize_semi_normal_ptn2(self):
        ins = cfm.sender_id_tlv(
            ma_domain=self.ma_domain,
            ma=self.ma,
        )
        buf = ins.serialize()
        form = '!BHBB2sB3s'
        res = struct.unpack_from(form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(8, res[1])
        eq_(0, res[2])
        eq_(self.ma_domain_length, res[3])
        eq_(self.ma_domain, res[4])
        eq_(self.ma_length, res[5])
        eq_(self.ma, res[6])

    def test_serialize_semi_normal_ptn3(self):
        ins = cfm.sender_id_tlv(
            chassis_id_subtype=self.chassis_id_subtype,
            chassis_id=self.chassis_id,
        )
        buf = ins.serialize()
        form = '!BHBB1sB'
        res = struct.unpack_from(form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(4, res[1])
        eq_(self.chassis_id_length, res[2])
        eq_(self.chassis_id_subtype, res[3])
        eq_(self.chassis_id, res[4])
        eq_(0, res[5])

    def test_serialize_semi_normal_ptn4(self):
        ins = cfm.sender_id_tlv(
            ma_domain=self.ma_domain,
        )
        buf = ins.serialize()
        form = '!BHBB2sB'
        res = struct.unpack_from(form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(5, res[1])
        eq_(0, res[2])
        eq_(self.ma_domain_length, res[3])
        eq_(self.ma_domain, res[4])
        eq_(0, res[5])

    def test_serialize_with_length_zero(self):
        ins = cfm.sender_id_tlv(
            0,
            0,
            self.chassis_id_subtype,
            self.chassis_id,
            0,
            self.ma_domain,
            0,
            self.ma,
        )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.chassis_id_length, res[2])
        eq_(self.chassis_id_subtype, res[3])
        eq_(self.chassis_id, res[4])
        eq_(self.ma_domain_length, res[5])
        eq_(self.ma_domain, res[6])
        eq_(self.ma_length, res[7])
        eq_(self.ma, res[8])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 10, len(self.ins))

    def test_default_args(self):
        ins = cfm.sender_id_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.sender_id_tlv._PACK_STR, six.binary_type(buf))
        eq_(res[0], cfm.CFM_SENDER_ID_TLV)
        eq_(res[1], 1)
        eq_(res[2], 0)


class Test_port_status_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_PORT_STATUS_TLV
        self.length = 1
        self.port_status = 1
        self.ins = cfm.port_status_tlv(
            self.length,
            self.port_status
        )
        self.form = '!BHB'
        self.buf = struct.pack(
            self.form,
            self._type,
            self.length,
            self.port_status
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.port_status, self.ins.port_status)

    def test_parser(self):
        _res = cfm.port_status_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.port_status, res.port_status)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.port_status, res[2])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 1, len(self.ins))

    def test_default_args(self):
        ins = cfm.port_status_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.port_status_tlv._PACK_STR, six.binary_type(buf))
        eq_(res[0], cfm.CFM_PORT_STATUS_TLV)
        eq_(res[1], 1)
        eq_(res[2], 2)


class Test_data_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_DATA_TLV
        self.length = 3
        self.data_value = b"\x01\x02\x03"
        self.ins = cfm.data_tlv(
            self.length,
            self.data_value
        )
        self.form = '!BH3s'
        self.buf = struct.pack(
            self.form,
            self._type,
            self.length,
            self.data_value,
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.data_value, self.ins.data_value)

    def test_parser(self):
        _res = cfm.data_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.data_value, res.data_value)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.data_value, res[2])

    def test_serialize_with_length_zero(self):
        ins = cfm.data_tlv(
            0,
            self.data_value
        )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.data_value, res[2])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 3, len(self.ins))

    def test_default_args(self):
        ins = cfm.data_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.data_tlv._PACK_STR, six.binary_type(buf))
        eq_(res[0], cfm.CFM_DATA_TLV)
        eq_(res[1], 0)


class Test_interface_status_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_INTERFACE_STATUS_TLV
        self.length = 1
        self.interface_status = 4
        self.ins = cfm.interface_status_tlv(
            self.length,
            self.interface_status
        )
        self.form = '!BHB'
        self.buf = struct.pack(
            self.form,
            self._type,
            self.length,
            self.interface_status
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.interface_status, self.ins.interface_status)

    def test_parser(self):
        _res = cfm.interface_status_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.interface_status, res.interface_status)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.interface_status, res[2])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 1, len(self.ins))

    def test_default_args(self):
        ins = cfm.interface_status_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.interface_status_tlv._PACK_STR, six.binary_type(buf))
        eq_(res[0], cfm.CFM_INTERFACE_STATUS_TLV)
        eq_(res[1], 1)
        eq_(res[2], 1)


class Test_ltm_egress_identifier_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_LTM_EGRESS_IDENTIFIER_TLV
        self.length = 8
        self.egress_id_ui = 7
        self.egress_id_mac = "11:22:33:44:55:66"
        self.ins = cfm.ltm_egress_identifier_tlv(
            self.length,
            self.egress_id_ui,
            self.egress_id_mac
        )
        self.form = '!BHH6s'
        self.buf = struct.pack(
            self.form,
            self._type,
            self.length,
            self.egress_id_ui,
            addrconv.mac.text_to_bin(self.egress_id_mac)
        )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.egress_id_ui, self.ins.egress_id_ui)
        eq_(self.egress_id_mac, self.ins.egress_id_mac)

    def test_parser(self):
        _res = cfm.ltm_egress_identifier_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.egress_id_ui, res.egress_id_ui)
        eq_(self.egress_id_mac, res.egress_id_mac)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.egress_id_ui, res[2])
        eq_(addrconv.mac.text_to_bin(self.egress_id_mac), res[3])

    def test_serialize_with_length_zero(self):
        ins = cfm.ltm_egress_identifier_tlv(
            0,
            self.egress_id_ui,
            self.egress_id_mac
        )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.egress_id_ui, res[2])
        eq_(addrconv.mac.text_to_bin(self.egress_id_mac), res[3])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 8, len(self.ins))

    def test_default_args(self):
        ins = cfm.ltm_egress_identifier_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(
            cfm.ltm_egress_identifier_tlv._PACK_STR, six.binary_type(buf))
        eq_(res[0], cfm.CFM_LTM_EGRESS_IDENTIFIER_TLV)
        eq_(res[1], 8)
        eq_(res[2], 0)
        eq_(res[3], addrconv.mac.text_to_bin('00:00:00:00:00:00'))


class Test_ltr_egress_identifier_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_LTR_EGRESS_IDENTIFIER_TLV
        self.length = 16
        self.last_egress_id_ui = 7
        self.last_egress_id_mac = "11:22:33:44:55:66"
        self.next_egress_id_ui = 5
        self.next_egress_id_mac = "33:11:33:aa:bb:cc"
        self.ins = cfm.ltr_egress_identifier_tlv(self.length,
                                                 self.last_egress_id_ui,
                                                 self.last_egress_id_mac,
                                                 self.next_egress_id_ui,
                                                 self.next_egress_id_mac
                                                 )
        self.form = '!BHH6sH6s'
        self.buf = struct.pack(
            self.form,
            self._type,
            self.length,
            self.last_egress_id_ui,
            addrconv.mac.text_to_bin(self.last_egress_id_mac),
            self.next_egress_id_ui,
            addrconv.mac.text_to_bin(self.next_egress_id_mac))

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.last_egress_id_ui, self.ins.last_egress_id_ui)
        eq_(self.last_egress_id_mac, self.ins.last_egress_id_mac)
        eq_(self.next_egress_id_ui, self.ins.next_egress_id_ui)
        eq_(self.next_egress_id_mac, self.ins.next_egress_id_mac)

    def test_parser(self):
        _res = cfm.ltr_egress_identifier_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.last_egress_id_ui, res.last_egress_id_ui)
        eq_(self.last_egress_id_mac, res.last_egress_id_mac)
        eq_(self.next_egress_id_ui, res.next_egress_id_ui)
        eq_(self.next_egress_id_mac, res.next_egress_id_mac)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.last_egress_id_ui, res[2])
        eq_(addrconv.mac.text_to_bin(self.last_egress_id_mac), res[3])
        eq_(self.next_egress_id_ui, res[4])
        eq_(addrconv.mac.text_to_bin(self.next_egress_id_mac), res[5])

    def test_serialize_with_length_zero(self):
        ins = cfm.ltr_egress_identifier_tlv(0,
                                            self.last_egress_id_ui,
                                            self.last_egress_id_mac,
                                            self.next_egress_id_ui,
                                            self.next_egress_id_mac
                                            )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.last_egress_id_ui, res[2])
        eq_(addrconv.mac.text_to_bin(self.last_egress_id_mac), res[3])
        eq_(self.next_egress_id_ui, res[4])
        eq_(addrconv.mac.text_to_bin(self.next_egress_id_mac), res[5])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 16, len(self.ins))

    def test_default_args(self):
        ins = cfm.ltr_egress_identifier_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.ltr_egress_identifier_tlv._PACK_STR,
                                 six.binary_type(buf))
        eq_(res[0], cfm.CFM_LTR_EGRESS_IDENTIFIER_TLV)
        eq_(res[1], 16)
        eq_(res[2], 0)
        eq_(res[3], addrconv.mac.text_to_bin('00:00:00:00:00:00'))
        eq_(res[4], 0)
        eq_(res[5], addrconv.mac.text_to_bin('00:00:00:00:00:00'))


class Test_organization_specific_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_ORGANIZATION_SPECIFIC_TLV
        self.length = 10
        self.oui = b"\xff\x12\x34"
        self.subtype = 3
        self.value = b"\x01\x02\x0f\x0e\x0d\x0c"
        self.ins = cfm.organization_specific_tlv(self.length,
                                                 self.oui,
                                                 self.subtype,
                                                 self.value
                                                 )
        self.form = '!BH3sB6s'
        self.buf = struct.pack(self.form,
                               self._type,
                               self.length,
                               self.oui,
                               self.subtype,
                               self.value
                               )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.oui, self.ins.oui)
        eq_(self.subtype, self.ins.subtype)
        eq_(self.value, self.ins.value)

    def test_parser(self):
        _res = cfm.organization_specific_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.oui, res.oui)
        eq_(self.subtype, res.subtype)
        eq_(self.value, res.value)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.oui, res[2])
        eq_(self.subtype, res[3])
        eq_(self.value, res[4])

    def test_serialize_with_zero(self):
        ins = cfm.organization_specific_tlv(0,
                                            self.oui,
                                            self.subtype,
                                            self.value
                                            )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.oui, res[2])
        eq_(self.subtype, res[3])
        eq_(self.value, res[4])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 10, len(self.ins))

    def test_default_args(self):
        ins = cfm.organization_specific_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.organization_specific_tlv._PACK_STR,
                                 six.binary_type(buf))
        eq_(res[0], cfm.CFM_ORGANIZATION_SPECIFIC_TLV)
        eq_(res[1], 4)
        eq_(res[2], b"\x00\x00\x00")
        eq_(res[3], 0)


class Test_reply_ingress_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_REPLY_INGRESS_TLV
        self.length = 12
        self.action = 2
        self.mac_address = 'aa:bb:cc:56:34:12'
        self.port_id_length = 3
        self.port_id_subtype = 2
        self.port_id = b"\x01\x04\x09"
        self.ins = cfm.reply_ingress_tlv(self.length, self.action,
                                         self.mac_address,
                                         self.port_id_length,
                                         self.port_id_subtype,
                                         self.port_id
                                         )
        self.form = '!BHB6sBB3s'
        self.buf = struct.pack(self.form,
                               self._type,
                               self.length,
                               self.action,
                               addrconv.mac.text_to_bin(self.mac_address),
                               self.port_id_length,
                               self.port_id_subtype,
                               self.port_id
                               )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.action, self.ins.action)
        eq_(self.mac_address, self.ins.mac_address)
        eq_(self.port_id_length, self.ins.port_id_length)
        eq_(self.port_id_subtype, self.ins.port_id_subtype)
        eq_(self.port_id, self.ins.port_id)

    def test_parser(self):
        _res = cfm.reply_ingress_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.action, res.action)
        eq_(self.mac_address, res.mac_address)
        eq_(self.port_id_length, res.port_id_length)
        eq_(self.port_id_subtype, res.port_id_subtype)
        eq_(self.port_id, res.port_id)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.action, res[2])
        eq_(addrconv.mac.text_to_bin(self.mac_address), res[3])
        eq_(self.port_id_length, res[4])
        eq_(self.port_id_subtype, res[5])
        eq_(self.port_id, res[6])

    def test_serialize_with_zero(self):
        ins = cfm.reply_ingress_tlv(0,
                                    self.action,
                                    self.mac_address,
                                    0,
                                    self.port_id_subtype,
                                    self.port_id
                                    )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.action, res[2])
        eq_(addrconv.mac.text_to_bin(self.mac_address), res[3])
        eq_(self.port_id_length, res[4])
        eq_(self.port_id_subtype, res[5])
        eq_(self.port_id, res[6])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 12, len(self.ins))

    def test_default_args(self):
        ins = cfm.reply_ingress_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.reply_ingress_tlv._PACK_STR, six.binary_type(buf))
        eq_(res[0], cfm.CFM_REPLY_INGRESS_TLV)
        eq_(res[1], 7)
        eq_(res[2], 1)
        eq_(res[3], addrconv.mac.text_to_bin('00:00:00:00:00:00'))


class Test_reply_egress_tlv(unittest.TestCase):

    def setUp(self):
        self._type = cfm.CFM_REPLY_EGRESS_TLV
        self.length = 12
        self.action = 2
        self.mac_address = 'aa:bb:cc:56:34:12'
        self.port_id_length = 3
        self.port_id_subtype = 2
        self.port_id = b"\x01\x04\x09"
        self.ins = cfm.reply_egress_tlv(self.length,
                                        self.action,
                                        self.mac_address,
                                        self.port_id_length,
                                        self.port_id_subtype,
                                        self.port_id
                                        )
        self.form = '!BHB6sBB3s'
        self.buf = struct.pack(self.form,
                               self._type,
                               self.length,
                               self.action,
                               addrconv.mac.text_to_bin(self.mac_address),
                               self.port_id_length,
                               self.port_id_subtype,
                               self.port_id
                               )

    def tearDown(self):
        pass

    def test_init(self):
        eq_(self.length, self.ins.length)
        eq_(self.action, self.ins.action)
        eq_(self.mac_address, self.ins.mac_address)
        eq_(self.port_id_length, self.ins.port_id_length)
        eq_(self.port_id_subtype, self.ins.port_id_subtype)
        eq_(self.port_id, self.ins.port_id)

    def test_parser(self):
        _res = cfm.reply_ingress_tlv.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        eq_(self.length, res.length)
        eq_(self.action, res.action)
        eq_(self.mac_address, res.mac_address)
        eq_(self.port_id_length, res.port_id_length)
        eq_(self.port_id_subtype, res.port_id_subtype)
        eq_(self.port_id, res.port_id)

    def test_serialize(self):
        buf = self.ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.action, res[2])
        eq_(addrconv.mac.text_to_bin(self.mac_address), res[3])
        eq_(self.port_id_length, res[4])
        eq_(self.port_id_subtype, res[5])
        eq_(self.port_id, res[6])

    def test_serialize_with_zero(self):
        ins = cfm.reply_egress_tlv(0,
                                   self.action,
                                   self.mac_address,
                                   0,
                                   self.port_id_subtype,
                                   self.port_id
                                   )
        buf = ins.serialize()
        res = struct.unpack_from(self.form, six.binary_type(buf))
        eq_(self._type, res[0])
        eq_(self.length, res[1])
        eq_(self.action, res[2])
        eq_(addrconv.mac.text_to_bin(self.mac_address), res[3])
        eq_(self.port_id_length, res[4])
        eq_(self.port_id_subtype, res[5])
        eq_(self.port_id, res[6])

    def test_len(self):
        # tlv_length = type_len + length_len + value_len
        eq_(1 + 2 + 12, len(self.ins))

    def test_default_args(self):
        ins = cfm.reply_egress_tlv()
        buf = ins.serialize()
        res = struct.unpack_from(cfm.reply_egress_tlv._PACK_STR,
                                 six.binary_type(buf))
        eq_(res[0], cfm.CFM_REPLY_EGRESS_TLV)
        eq_(res[1], 7)
        eq_(res[2], 1)
        eq_(res[3], addrconv.mac.text_to_bin('00:00:00:00:00:00'))



# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Defines data types and models required specifically for EVPN support.
"""

import logging

from os_ken.lib.packet.bgp import EvpnNLRI
from os_ken.lib.packet.bgp import RF_L2_EVPN

from os_ken.services.protocols.bgp.info_base.vpn import VpnDest
from os_ken.services.protocols.bgp.info_base.vpn import VpnPath
from os_ken.services.protocols.bgp.info_base.vpn import VpnTable

LOG = logging.getLogger('bgpspeaker.info_base.evpn')


class EvpnDest(VpnDest):
    """EVPN Destination

    Store EVPN Paths.
    """
    ROUTE_FAMILY = RF_L2_EVPN


class EvpnTable(VpnTable):
    """Global table to store EVPN routing information.

    Uses `EvpnDest` to store destination information for each known EVPN
    paths.
    """
    ROUTE_FAMILY = RF_L2_EVPN
    VPN_DEST_CLASS = EvpnDest


class EvpnPath(VpnPath):
    """Represents a way of reaching an EVPN destination."""
    ROUTE_FAMILY = RF_L2_EVPN
    VRF_PATH_CLASS = None  # defined in init - anti cyclic import hack
    NLRI_CLASS = EvpnNLRI

    def __init__(self, *args, **kwargs):
        super(EvpnPath, self).__init__(*args, **kwargs)
        from os_ken.services.protocols.bgp.info_base.vrfevpn import VrfEvpnPath
        self.VRF_PATH_CLASS = VrfEvpnPath



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import six
import struct
from . import packet_base
from . import icmpv6
from . import tcp
from . import udp
from . import sctp
from . import gre
from . import in_proto as inet
from os_ken.lib import addrconv
from os_ken.lib import stringify


IPV6_ADDRESS_PACK_STR = '!16s'
IPV6_ADDRESS_LEN = struct.calcsize(IPV6_ADDRESS_PACK_STR)
IPV6_PSEUDO_HEADER_PACK_STR = '!16s16s3xB'


class ipv6(packet_base.PacketBase):
    """IPv6 (RFC 2460) header encoder/decoder class.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    IPv6 addresses are represented as a string like 'ff02::1'.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|p{30em}|l|

    ============== ======================================== ==================
    Attribute      Description                              Example
    ============== ======================================== ==================
    version        Version
    traffic_class  Traffic Class
    flow_label     When decoding, Flow Label.
                   When encoding, the most significant 8
                   bits of Flow Label.
    payload_length Payload Length
    nxt            Next Header
    hop_limit      Hop Limit
    src            Source Address                           'ff02::1'
    dst            Destination Address                      '::'
    ext_hdrs       Extension Headers
    ============== ======================================== ==================
    """

    _PACK_STR = '!IHBB16s16s'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _IPV6_EXT_HEADER_TYPE = {}
    _TYPE = {
        'ascii': [
            'src', 'dst'
        ]
    }

    @staticmethod
    def register_header_type(type_):
        def _register_header_type(cls):
            ipv6._IPV6_EXT_HEADER_TYPE[type_] = cls
            return cls
        return _register_header_type

    def __init__(self, version=6, traffic_class=0, flow_label=0,
                 payload_length=0, nxt=inet.IPPROTO_TCP, hop_limit=255,
                 src='10::10', dst='20::20', ext_hdrs=None):
        super(ipv6, self).__init__()
        self.version = version
        self.traffic_class = traffic_class
        self.flow_label = flow_label
        self.payload_length = payload_length
        self.nxt = nxt
        self.hop_limit = hop_limit
        self.src = src
        self.dst = dst
        ext_hdrs = ext_hdrs or []
        assert isinstance(ext_hdrs, list)
        for ext_hdr in ext_hdrs:
            assert isinstance(ext_hdr, header)
        self.ext_hdrs = ext_hdrs

    @classmethod
    def parser(cls, buf):
        (v_tc_flow, payload_length, nxt, hlim, src, dst) = struct.unpack_from(
            cls._PACK_STR, buf)
        version = v_tc_flow >> 28
        traffic_class = (v_tc_flow >> 20) & 0xff
        flow_label = v_tc_flow & 0xfffff
        hop_limit = hlim
        offset = cls._MIN_LEN
        last = nxt
        ext_hdrs = []
        while True:
            cls_ = cls._IPV6_EXT_HEADER_TYPE.get(last)
            if not cls_:
                break
            hdr = cls_.parser(buf[offset:])
            ext_hdrs.append(hdr)
            offset += len(hdr)
            last = hdr.nxt
        msg = cls(version, traffic_class, flow_label, payload_length,
                  nxt, hop_limit, addrconv.ipv6.bin_to_text(src),
                  addrconv.ipv6.bin_to_text(dst), ext_hdrs)
        return (msg, ipv6.get_packet_type(last),
                buf[offset:offset + payload_length])

    def serialize(self, payload, prev):
        hdr = bytearray(40)
        v_tc_flow = (self.version << 28 | self.traffic_class << 20 |
                     self.flow_label)
        struct.pack_into(ipv6._PACK_STR, hdr, 0, v_tc_flow,
                         self.payload_length, self.nxt, self.hop_limit,
                         addrconv.ipv6.text_to_bin(self.src),
                         addrconv.ipv6.text_to_bin(self.dst))
        if self.ext_hdrs:
            for ext_hdr in self.ext_hdrs:
                hdr.extend(ext_hdr.serialize())
        if 0 == self.payload_length:
            payload_length = len(payload)
            for ext_hdr in self.ext_hdrs:
                payload_length += len(ext_hdr)
            self.payload_length = payload_length
            struct.pack_into('!H', hdr, 4, self.payload_length)
        return hdr

    def __len__(self):
        ext_hdrs_len = 0
        for ext_hdr in self.ext_hdrs:
            ext_hdrs_len += len(ext_hdr)
        return self._MIN_LEN + ext_hdrs_len


ipv6.register_packet_type(icmpv6.icmpv6, inet.IPPROTO_ICMPV6)
ipv6.register_packet_type(tcp.tcp, inet.IPPROTO_TCP)
ipv6.register_packet_type(udp.udp, inet.IPPROTO_UDP)
ipv6.register_packet_type(sctp.sctp, inet.IPPROTO_SCTP)
ipv6.register_packet_type(gre.gre, inet.IPPROTO_GRE)


@six.add_metaclass(abc.ABCMeta)
class header(stringify.StringifyMixin):
    """extension header abstract class."""

    def __init__(self, nxt):
        self.nxt = nxt

    @classmethod
    @abc.abstractmethod
    def parser(cls, buf):
        pass

    @abc.abstractmethod
    def serialize(self):
        pass

    @abc.abstractmethod
    def __len__(self):
        pass


class opt_header(header):
    """an abstract class for Hop-by-Hop Options header and destination
    header."""

    _PACK_STR = '!BB'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _FIX_SIZE = 8
    _class_prefixes = ['option']

    @abc.abstractmethod
    def __init__(self, nxt, size, data):
        super(opt_header, self).__init__(nxt)
        assert not (size % 8)
        self.size = size
        self.data = data

    @classmethod
    def parser(cls, buf):
        (nxt, len_) = struct.unpack_from(cls._PACK_STR, buf)
        data_len = cls._FIX_SIZE + int(len_)
        data = []
        size = cls._MIN_LEN
        while size < data_len:
            (type_, ) = struct.unpack_from('!B', buf[size:])
            if type_ == 0:
                opt = option(type_, -1, None)
                size += 1
            else:
                opt = option.parser(buf[size:])
                size += len(opt)
            data.append(opt)
        return cls(nxt, len_, data)

    def serialize(self):
        buf = struct.pack(self._PACK_STR, self.nxt, self.size)
        buf = bytearray(buf)
        if self.data is None:
            self.data = [option(type_=1, len_=4,
                                data=b'\x00\x00\x00\x00')]
        for opt in self.data:
            buf.extend(opt.serialize())
        return buf

    def __len__(self):
        return self._FIX_SIZE + self.size


@ipv6.register_header_type(inet.IPPROTO_HOPOPTS)
class hop_opts(opt_header):
    """IPv6 (RFC 2460) Hop-by-Hop Options header encoder/decoder class.

    This is used with os_ken.lib.packet.ipv6.ipv6.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    nxt            Next Header
    size           the length of the Hop-by-Hop Options header,
                   not include the first 8 octet.
    data           IPv6 options.
    ============== =======================================
    """
    TYPE = inet.IPPROTO_HOPOPTS

    def __init__(self, nxt=inet.IPPROTO_TCP, size=0, data=None):
        super(hop_opts, self).__init__(nxt, size, data)


@ipv6.register_header_type(inet.IPPROTO_DSTOPTS)
class dst_opts(opt_header):
    """IPv6 (RFC 2460) destination header encoder/decoder class.

    This is used with os_ken.lib.packet.ipv6.ipv6.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    nxt            Next Header
    size           the length of the destination header,
                   not include the first 8 octet.
    data           IPv6 options.
    ============== =======================================
    """
    TYPE = inet.IPPROTO_DSTOPTS

    def __init__(self, nxt=inet.IPPROTO_TCP, size=0, data=None):
        super(dst_opts, self).__init__(nxt, size, data)


class option(stringify.StringifyMixin):
    r"""IPv6 (RFC 2460) Options header encoder/decoder class.

    This is used with os_ken.lib.packet.ipv6.hop_opts or
                      os_ken.lib.packet.ipv6.dst_opts.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    type\_         option type.
    len\_          the length of data. -1 if type\_ is 0.
    data           an option value. None if len\_ is 0 or -1.
    ============== =======================================
    """

    _PACK_STR = '!BB'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, type_=0, len_=-1, data=None):
        self.type_ = type_
        self.len_ = len_
        self.data = data

    @classmethod
    def parser(cls, buf):
        (type_, ) = struct.unpack_from('!B', buf)
        if not type_:
            cls_ = cls(type_, -1, None)
        else:
            data = None
            (type_, len_) = struct.unpack_from(cls._PACK_STR, buf)
            if len_:
                form = "%ds" % len_
                (data, ) = struct.unpack_from(form, buf, cls._MIN_LEN)
            cls_ = cls(type_, len_, data)
        return cls_

    def serialize(self):
        data = None
        if not self.type_:
            data = struct.pack('!B', self.type_)
        elif not self.len_:
            data = struct.pack(self._PACK_STR, self.type_, self.len_)
        else:
            form = "%ds" % self.len_
            data = struct.pack(self._PACK_STR + form, self.type_,
                               self.len_, self.data)
        return data

    def __len__(self):
        return self._MIN_LEN + self.len_


@ipv6.register_header_type(inet.IPPROTO_ROUTING)
class routing(header):
    """An IPv6 Routing Header decoder class.
    This class has only the parser method.

    IPv6 Routing Header types.

    http://www.iana.org/assignments/ipv6-parameters/ipv6-parameters.xhtml

    +-----------+----------------------------------+-------------------+
    | Value     | Description                      | Reference         |
    +===========+==================================+===================+
    | 0         | Source Route (DEPRECATED)        | [[IPV6]][RFC5095] |
    +-----------+----------------------------------+-------------------+
    | 1         | Nimrod (DEPRECATED 2009-05-06)   |                   |
    +-----------+----------------------------------+-------------------+
    | 2         | Type 2 Routing Header            | [RFC6275]         |
    +-----------+----------------------------------+-------------------+
    | 3         | RPL Source Route Header          | [RFC6554]         |
    +-----------+----------------------------------+-------------------+
    | 4 - 252   | Unassigned                       |                   |
    +-----------+----------------------------------+-------------------+
    | 253       | RFC3692-style Experiment 1 [2]   | [RFC4727]         |
    +-----------+----------------------------------+-------------------+
    | 254       | RFC3692-style Experiment 2 [2]   | [RFC4727]         |
    +-----------+----------------------------------+-------------------+
    | 255       | Reserved                         |                   |
    +-----------+----------------------------------+-------------------+
    """

    TYPE = inet.IPPROTO_ROUTING

    _OFFSET_LEN = struct.calcsize('!2B')

    # IPv6 Routing Header Type
    ROUTING_TYPE_2 = 0x02
    ROUTING_TYPE_3 = 0x03

    @classmethod
    def parser(cls, buf):
        (type_, ) = struct.unpack_from('!B', buf, cls._OFFSET_LEN)
        switch = {
            # TODO: make parsers of type2.
            cls.ROUTING_TYPE_2: None,
            cls.ROUTING_TYPE_3: routing_type3
        }
        cls_ = switch.get(type_)
        if cls_:
            return cls_.parser(buf)
        else:
            return None


class routing_type3(header):
    """
    An IPv6 Routing Header for Source Routes with the RPL (RFC 6554)
    encoder/decoder class.

    This is used with os_ken.lib.packet.ipv6.ipv6.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    nxt            Next Header
    size           The length of the Routing header,
                   not include the first 8 octet.
                   (0 means automatically-calculate when encoding)
    type           Identifies the particular Routing header variant.
    seg            Number of route segments remaining.
    cmpi           Number of prefix octets from segments 1 through n-1.
    cmpe           Number of prefix octets from segment n.
    pad            Number of octets that are used for padding
                   after Address[n] at the end of the SRH.
    adrs           Vector of addresses, numbered 1 to n.
    ============== =======================================
    """

    _PACK_STR = '!BBBBBB2x'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _TYPE = {
        'asciilist': [
            'adrs'
        ]
    }

    def __init__(self, nxt=inet.IPPROTO_TCP, size=0,
                 type_=3, seg=0, cmpi=0, cmpe=0, adrs=None):
        super(routing_type3, self).__init__(nxt)
        self.size = size
        self.type_ = type_
        self.seg = seg
        self.cmpi = cmpi
        self.cmpe = cmpe
        adrs = adrs or []
        assert isinstance(adrs, list)
        self.adrs = adrs
        self._pad = (8 - ((len(self.adrs) - 1) * (16 - self.cmpi) +
                          (16 - self.cmpe) % 8)) % 8

    @classmethod
    def _get_size(cls, size):
        return (int(size) + 1) * 8

    @classmethod
    def parser(cls, buf):
        (nxt, size, type_, seg, cmp_, pad) = struct.unpack_from(
            cls._PACK_STR, buf)
        data = cls._MIN_LEN
        header_len = cls._get_size(size)
        cmpi = int(cmp_ >> 4)
        cmpe = int(cmp_ & 0xf)
        pad = int(pad >> 4)
        adrs = []
        if size:
            # Address[1..n-1] has size (16 - CmprI) octets
            adrs_len_i = 16 - cmpi
            # Address[n] has size (16 - CmprE) octets
            adrs_len_e = 16 - cmpe
            form_i = "%ds" % adrs_len_i
            form_e = "%ds" % adrs_len_e
            while data < (header_len - (adrs_len_e + pad)):
                (adr, ) = struct.unpack_from(form_i, buf[data:])
                adr = (b'\x00' * cmpi) + adr
                adrs.append(addrconv.ipv6.bin_to_text(adr))
                data += adrs_len_i
            (adr, ) = struct.unpack_from(form_e, buf[data:])
            adr = (b'\x00' * cmpe) + adr
            adrs.append(addrconv.ipv6.bin_to_text(adr))
        return cls(nxt, size, type_, seg, cmpi, cmpe, adrs)

    def serialize(self):
        if self.size == 0:
            self.size = ((len(self.adrs) - 1) * (16 - self.cmpi) +
                         (16 - self.cmpe) + self._pad) // 8
        buf = struct.pack(self._PACK_STR, self.nxt, self.size,
                          self.type_, self.seg, (self.cmpi << 4) | self.cmpe,
                          self._pad << 4)
        buf = bytearray(buf)
        if self.size:
            form_i = "%ds" % (16 - self.cmpi)
            form_e = "%ds" % (16 - self.cmpe)
            slice_i = slice(self.cmpi, 16)
            slice_e = slice(self.cmpe, 16)
            for adr in self.adrs[:-1]:
                buf.extend(
                    struct.pack(
                        form_i, addrconv.ipv6.text_to_bin(adr)[slice_i]))
            buf.extend(struct.pack(
                form_e,
                addrconv.ipv6.text_to_bin(self.adrs[-1])[slice_e]))
        return buf

    def __len__(self):
        return routing_type3._get_size(self.size)


@ipv6.register_header_type(inet.IPPROTO_FRAGMENT)
class fragment(header):
    r"""IPv6 (RFC 2460) fragment header encoder/decoder class.

    This is used with os_ken.lib.packet.ipv6.ipv6.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    nxt            Next Header
    offset         offset, in 8-octet units, relative to
                   the start of the fragmentable part of
                   the original packet.
    more           1 means more fragments follow;
                   0 means last fragment.
    id\_           packet identification value.
    ============== =======================================
    """
    TYPE = inet.IPPROTO_FRAGMENT

    _PACK_STR = '!BxHI'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, nxt=inet.IPPROTO_TCP, offset=0, more=0, id_=0):
        super(fragment, self).__init__(nxt)
        self.offset = offset
        self.more = more
        self.id_ = id_

    @classmethod
    def parser(cls, buf):
        (nxt, off_m, id_) = struct.unpack_from(cls._PACK_STR, buf)
        offset = off_m >> 3
        more = off_m & 0x1
        return cls(nxt, offset, more, id_)

    def serialize(self):
        off_m = (self.offset << 3 | self.more)
        buf = struct.pack(self._PACK_STR, self.nxt, off_m, self.id_)
        return buf

    def __len__(self):
        return self._MIN_LEN


@ipv6.register_header_type(inet.IPPROTO_AH)
class auth(header):
    """IP Authentication header (RFC 2402) encoder/decoder class.

    This is used with os_ken.lib.packet.ipv6.ipv6.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    nxt            Next Header
    size           the length of the Authentication Header
                   in 64-bit words, subtracting 1.
    spi            security parameters index.
    seq            sequence number.
    data           authentication data.
    ============== =======================================
    """
    TYPE = inet.IPPROTO_AH

    _PACK_STR = '!BB2xII'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, nxt=inet.IPPROTO_TCP, size=2, spi=0, seq=0,
                 data=b'\x00\x00\x00\x00'):
        super(auth, self).__init__(nxt)
        assert data is not None
        self.size = size
        self.spi = spi
        self.seq = seq
        self.data = data

    @classmethod
    def _get_size(cls, size):
        return (int(size) + 2) * 4

    @classmethod
    def parser(cls, buf):
        (nxt, size, spi, seq) = struct.unpack_from(cls._PACK_STR, buf)
        form = "%ds" % (cls._get_size(size) - cls._MIN_LEN)
        (data, ) = struct.unpack_from(form, buf, cls._MIN_LEN)
        return cls(nxt, size, spi, seq, data)

    def serialize(self):
        buf = struct.pack(self._PACK_STR, self.nxt, self.size, self.spi,
                          self.seq)
        buf = bytearray(buf)
        form = "%ds" % (auth._get_size(self.size) - self._MIN_LEN)
        buf.extend(struct.pack(form, self.data))
        return buf

    def __len__(self):
        return auth._get_size(self.size)


ipv6.set_classes(ipv6._IPV6_EXT_HEADER_TYPE)



#!/usr/bin/env python


import os
import sys

from nose import config
from nose import core

sys.path.append(os.getcwd())
sys.path.append(os.path.dirname(__file__))


import os_ken.tests.unit
from os_ken.tests.test_lib import run_tests


if __name__ == '__main__':
    exit_status = False

    # if a single test case was specified,
    # we should only invoked the tests once
    invoke_once = len(sys.argv) > 1

    cwd = os.getcwd()
    c = config.Config(stream=sys.stdout,
                      env=os.environ,
                      verbosity=int(os.environ.get('NOSE_VERBOSE', 3)),
                      includeExe=True,
                      traverseNamespace=True,
                      plugins=core.DefaultPluginManager())
    c.configureWhere(os_ken.tests.unit.__path__)

    exit_status = run_tests(c)
    sys.exit(exit_status)




from os_ken.services.protocols.bgp.base import ActivityException
from os_ken.services.protocols.bgp.operator.command import Command
from os_ken.services.protocols.bgp.operator.command import CommandsResponse
from os_ken.services.protocols.bgp.operator.command import STATUS_ERROR
from os_ken.services.protocols.bgp.operator.command import STATUS_OK
from os_ken.services.protocols.bgp.operator.commands.responses import (
    WrongParamResp)
from .route_formatter_mixin import RouteFormatterMixin


class RibBase(Command, RouteFormatterMixin):
    supported_families = [
        'ipv4',
        'ipv6',
        'vpnv4',
        'vpnv6',
        'rtfilter',
        'evpn',
        'ipv4fs',
        'ipv6fs',
        'vpnv4fs',
        'vpnv6fs',
        'l2vpnfs',
    ]


class Rib(RibBase):
    help_msg = 'show all routes for address family'
    param_help_msg = '<address-family>'
    command = 'rib'

    def __init__(self, *args, **kwargs):
        super(Rib, self).__init__(*args, **kwargs)
        self.subcommands = {
            'all': self.All}

    def action(self, params):
        if len(params) != 1 or params[0] not in self.supported_families:
            return WrongParamResp()
        from os_ken.services.protocols.bgp.operator.internal_api \
            import WrongParamError
        try:
            return CommandsResponse(
                STATUS_OK,
                self.api.get_single_rib_routes(params[0])
            )
        except WrongParamError as e:
            return WrongParamResp(e)

    @classmethod
    def cli_resp_formatter(cls, resp):
        if resp.status == STATUS_ERROR:
            return RibBase.cli_resp_formatter(resp)
        return cls._format_family_header() + cls._format_family(resp.value)

    class All(RibBase):
        help_msg = 'show routes for all RIBs'
        command = 'all'

        def action(self, params):
            if len(params) != 0:
                return WrongParamResp()
            ret = {}
            try:
                for family in self.supported_families:
                    ret[family] = self.api.get_single_rib_routes(family)
                return CommandsResponse(STATUS_OK, ret)
            except ActivityException as e:
                return CommandsResponse(STATUS_ERROR, e)

        @classmethod
        def cli_resp_formatter(cls, resp):
            if resp.status == STATUS_ERROR:
                return RibBase.cli_resp_formatter(resp)
            ret = cls._format_family_header()
            for family, data in resp.value.items():
                ret += 'Family: {0}\n'.format(family)
                ret += cls._format_family(data)
            return ret



# Copyright (C) 2011-2014 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2011 Isaku Yamahata <yamahata at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The central management of OSKen applications.

- Load OSKen applications
- Provide `contexts` to OSKen applications
- Route messages among OSKen applications

"""

import inspect
import itertools
import logging
import sys
import os
import gc

from os_ken import cfg
from os_ken import utils
from os_ken.controller.handler import register_instance, get_dependent_services
from os_ken.controller.controller import Datapath
from os_ken.controller import event
from os_ken.controller.event import EventRequestBase, EventReplyBase
from os_ken.lib import hub
from os_ken.ofproto import ofproto_protocol

LOG = logging.getLogger('os_ken.base.app_manager')

SERVICE_BRICKS = {}


def lookup_service_brick(name):
    return SERVICE_BRICKS.get(name)


def _lookup_service_brick_by_ev_cls(ev_cls):
    return _lookup_service_brick_by_mod_name(ev_cls.__module__)


def _lookup_service_brick_by_mod_name(mod_name):
    return lookup_service_brick(mod_name.split('.')[-1])


def register_app(app):
    assert isinstance(app, OSKenApp)
    assert app.name not in SERVICE_BRICKS
    SERVICE_BRICKS[app.name] = app
    register_instance(app)


def unregister_app(app):
    SERVICE_BRICKS.pop(app.name)


def require_app(app_name, api_style=False):
    """
    Request the application to be automatically loaded.

    If this is used for "api" style modules, which is imported by a client
    application, set api_style=True.

    If this is used for client application module, set api_style=False.
    """
    iterable = (inspect.getmodule(frame[0]) for frame in inspect.stack())
    modules = [module for module in iterable if module is not None]
    if api_style:
        m = modules[2]  # skip a frame for "api" module
    else:
        m = modules[1]
    m._REQUIRED_APP = getattr(m, '_REQUIRED_APP', [])
    m._REQUIRED_APP.append(app_name)
    LOG.debug('require_app: %s is required by %s', app_name, m.__name__)


class OSKenApp(object):
    """
    The base class for OSKen applications.

    OSKenApp subclasses are instantiated after osken-manager loaded
    all requested OSKen application modules.
    __init__ should call OSKenApp.__init__ with the same arguments.
    It's illegal to send any events in __init__.

    The instance attribute 'name' is the name of the class used for
    message routing among OSKen applications.  (Cf. send_event)
    It's set to __class__.__name__ by OSKenApp.__init__.
    It's discouraged for subclasses to override this.
    """

    _CONTEXTS = {}
    """
    A dictionary to specify contexts which this OSKen application wants to use.
    Its key is a name of context and its value is an ordinary class
    which implements the context.  The class is instantiated by app_manager
    and the instance is shared among OSKenApp subclasses which has _CONTEXTS
    member with the same key.  A OSKenApp subclass can obtain a reference to
    the instance via its __init__'s kwargs as the following.

    Example::

        _CONTEXTS = {
            'network': network.Network
        }

        def __init__(self, *args, *kwargs):
            self.network = kwargs['network']
    """

    _EVENTS = []
    """
    A list of event classes which this OSKenApp subclass would generate.
    This should be specified if and only if event classes are defined in
    a different python module from the OSKenApp subclass is.
    """

    OFP_VERSIONS = None
    """
    A list of supported OpenFlow versions for this OSKenApp.
    The default is all versions supported by the framework.

    Examples::

        OFP_VERSIONS = [ofproto_v1_0.OFP_VERSION,
                        ofproto_v1_2.OFP_VERSION]

    If multiple OSKen applications are loaded in the system,
    the intersection of their OFP_VERSIONS is used.
    """

    @classmethod
    def context_iteritems(cls):
        """
        Return iterator over the (key, contxt class) of application context
        """
        return iter(cls._CONTEXTS.items())

    def __init__(self, *_args, **_kwargs):
        super(OSKenApp, self).__init__()
        self.name = self.__class__.__name__
        self.event_handlers = {}        # ev_cls -> handlers:list
        self.observers = {}     # ev_cls -> observer-name -> states:set
        self.threads = []
        self.main_thread = None
        self.events = hub.Queue(128)
        self._events_sem = hub.BoundedSemaphore(self.events.maxsize)
        if hasattr(self.__class__, 'LOGGER_NAME'):
            self.logger = logging.getLogger(self.__class__.LOGGER_NAME)
        else:
            self.logger = logging.getLogger(self.name)
        self.CONF = cfg.CONF

        # prevent accidental creation of instances of this class outside OSKenApp
        class _EventThreadStop(event.EventBase):
            pass
        self._event_stop = _EventThreadStop()
        self.is_active = True

    def start(self):
        """
        Hook that is called after startup initialization is done.
        """
        self.threads.append(hub.spawn(self._event_loop))

    def stop(self):
        if self.main_thread:
            hub.kill(self.main_thread)
        self.is_active = False
        self._send_event(self._event_stop, None)
        hub.joinall(self.threads)

    def set_main_thread(self, thread):
        """
        Set self.main_thread so that stop() can terminate it.

        Only AppManager.instantiate_apps should call this function.
        """
        self.main_thread = thread

    def register_handler(self, ev_cls, handler):
        assert callable(handler)
        self.event_handlers.setdefault(ev_cls, [])
        self.event_handlers[ev_cls].append(handler)

    def unregister_handler(self, ev_cls, handler):
        assert callable(handler)
        self.event_handlers[ev_cls].remove(handler)
        if not self.event_handlers[ev_cls]:
            del self.event_handlers[ev_cls]

    def register_observer(self, ev_cls, name, states=None):
        states = states or set()
        ev_cls_observers = self.observers.setdefault(ev_cls, {})
        ev_cls_observers.setdefault(name, set()).update(states)

    def unregister_observer(self, ev_cls, name):
        observers = self.observers.get(ev_cls, {})
        observers.pop(name)

    def unregister_observer_all_event(self, name):
        for observers in self.observers.values():
            observers.pop(name, None)

    def observe_event(self, ev_cls, states=None):
        brick = _lookup_service_brick_by_ev_cls(ev_cls)
        if brick is not None:
            brick.register_observer(ev_cls, self.name, states)

    def unobserve_event(self, ev_cls):
        brick = _lookup_service_brick_by_ev_cls(ev_cls)
        if brick is not None:
            brick.unregister_observer(ev_cls, self.name)

    def get_handlers(self, ev, state=None):
        """Returns a list of handlers for the specific event.

        :param ev: The event to handle.
        :param state: The current state. ("dispatcher")
                      If None is given, returns all handlers for the event.
                      Otherwise, returns only handlers that are interested
                      in the specified state.
                      The default is None.
        """
        ev_cls = ev.__class__
        handlers = self.event_handlers.get(ev_cls, [])
        if state is None:
            return handlers

        def test(h):
            if not hasattr(h, 'callers') or ev_cls not in h.callers:
                # dynamically registered handlers does not have
                # h.callers element for the event.
                return True
            states = h.callers[ev_cls].dispatchers
            if not states:
                # empty states means all states
                return True
            return state in states

        return filter(test, handlers)

    def get_observers(self, ev, state):
        observers = []
        for k, v in self.observers.get(ev.__class__, {}).items():
            if not state or not v or state in v:
                observers.append(k)

        return observers

    def send_request(self, req):
        """
        Make a synchronous request.
        Set req.sync to True, send it to a OSKen application specified by
        req.dst, and block until receiving a reply.
        Returns the received reply.
        The argument should be an instance of EventRequestBase.
        """

        assert isinstance(req, EventRequestBase)
        req.sync = True
        req.reply_q = hub.Queue()
        self.send_event(req.dst, req)
        # going to sleep for the reply
        return req.reply_q.get()

    def _event_loop(self):
        while self.is_active or not self.events.empty():
            ev, state = self.events.get()
            self._events_sem.release()
            if ev == self._event_stop:
                continue
            handlers = self.get_handlers(ev, state)
            for handler in handlers:
                try:
                    handler(ev)
                except hub.TaskExit:
                    # Normal exit.
                    # Propagate upwards, so we leave the event loop.
                    raise
                except:
                    LOG.exception('%s: Exception occurred during handler processing. '
                                  'Backtrace from offending handler '
                                  '[%s] servicing event [%s] follows.',
                                  self.name, handler.__name__, ev.__class__.__name__)

    def _send_event(self, ev, state):
        self._events_sem.acquire()
        self.events.put((ev, state))

    def send_event(self, name, ev, state=None):
        """
        Send the specified event to the OSKenApp instance specified by name.
        """

        if name in SERVICE_BRICKS:
            if isinstance(ev, EventRequestBase):
                ev.src = self.name
            LOG.debug("EVENT %s->%s %s",
                      self.name, name, ev.__class__.__name__)
            SERVICE_BRICKS[name]._send_event(ev, state)
        else:
            LOG.debug("EVENT LOST %s->%s %s",
                      self.name, name, ev.__class__.__name__)

    def send_event_to_observers(self, ev, state=None):
        """
        Send the specified event to all observers of this OSKenApp.
        """

        for observer in self.get_observers(ev, state):
            self.send_event(observer, ev, state)

    def reply_to_request(self, req, rep):
        """
        Send a reply for a synchronous request sent by send_request.
        The first argument should be an instance of EventRequestBase.
        The second argument should be an instance of EventReplyBase.
        """

        assert isinstance(req, EventRequestBase)
        assert isinstance(rep, EventReplyBase)
        rep.dst = req.src
        if req.sync:
            req.reply_q.put(rep)
        else:
            self.send_event(rep.dst, rep)

    def close(self):
        """
        teardown method.
        The method name, close, is chosen for python context manager
        """
        pass


class AppManager(object):
    # singleton
    _instance = None

    @staticmethod
    def run_apps(app_lists):
        """Run a set of OSKen applications

        A convenient method to load and instantiate apps.
        This blocks until all relevant apps stop.
        """
        app_mgr = AppManager.get_instance()
        app_mgr.load_apps(app_lists)
        contexts = app_mgr.create_contexts()
        services = app_mgr.instantiate_apps(**contexts)
        try:
            hub.joinall(services)
        finally:
            app_mgr.close()
            for t in services:
                t.kill()
            hub.joinall(services)
            gc.collect()

    @staticmethod
    def get_instance():
        if not AppManager._instance:
            AppManager._instance = AppManager()
        return AppManager._instance

    def __init__(self):
        self.applications_cls = {}
        self.applications = {}
        self.contexts_cls = {}
        self.contexts = {}
        self.close_sem = hub.Semaphore()

    def load_app(self, name):
        mod = utils.import_module(name)
        clses = inspect.getmembers(mod,
                                   lambda cls: (inspect.isclass(cls) and
                                                issubclass(cls, OSKenApp) and
                                                mod.__name__ ==
                                                cls.__module__))
        if clses:
            return clses[0][1]
        return None

    def load_apps(self, app_lists):
        app_lists = [app for app
                     in itertools.chain.from_iterable(app.split(',')
                                                      for app in app_lists)]
        while len(app_lists) > 0:
            app_cls_name = app_lists.pop(0)

            context_modules = [x.__module__ for x in self.contexts_cls.values()]
            if app_cls_name in context_modules:
                continue

            LOG.info('loading app %s', app_cls_name)

            cls = self.load_app(app_cls_name)
            if cls is None:
                continue

            self.applications_cls[app_cls_name] = cls

            services = []
            for key, context_cls in cls.context_iteritems():
                v = self.contexts_cls.setdefault(key, context_cls)
                assert v == context_cls
                context_modules.append(context_cls.__module__)

                if issubclass(context_cls, OSKenApp):
                    services.extend(get_dependent_services(context_cls))

            # we can't load an app that will be initiataed for
            # contexts.
            for i in get_dependent_services(cls):
                if i not in context_modules:
                    services.append(i)
            if services:
                app_lists.extend([s for s in set(services)
                                  if s not in app_lists])

    def create_contexts(self):
        for key, cls in self.contexts_cls.items():
            if issubclass(cls, OSKenApp):
                # hack for dpset
                context = self._instantiate(None, cls)
            else:
                context = cls()
            LOG.info('creating context %s', key)
            assert key not in self.contexts
            self.contexts[key] = context
        return self.contexts

    def _update_bricks(self):
        for i in SERVICE_BRICKS.values():
            for _k, m in inspect.getmembers(i, inspect.ismethod):
                if not hasattr(m, 'callers'):
                    continue
                for ev_cls, c in m.callers.items():
                    if not c.ev_source:
                        continue

                    brick = _lookup_service_brick_by_mod_name(c.ev_source)
                    if brick:
                        brick.register_observer(ev_cls, i.name,
                                                c.dispatchers)

                    # allow OSKenApp and Event class are in different module
                    for brick in SERVICE_BRICKS.values():
                        if ev_cls in brick._EVENTS:
                            brick.register_observer(ev_cls, i.name,
                                                    c.dispatchers)

    @staticmethod
    def _report_brick(name, app):
        LOG.debug("BRICK %s", name)
        for ev_cls, list_ in app.observers.items():
            LOG.debug("  PROVIDES %s TO %s", ev_cls.__name__, list_)
        for ev_cls in app.event_handlers.keys():
            LOG.debug("  CONSUMES %s", ev_cls.__name__)

    @staticmethod
    def report_bricks():
        for brick, i in SERVICE_BRICKS.items():
            AppManager._report_brick(brick, i)

    def _instantiate(self, app_name, cls, *args, **kwargs):
        # for now, only single instance of a given module
        # Do we need to support multiple instances?
        # Yes, maybe for slicing.
        LOG.info('instantiating app %s of %s', app_name, cls.__name__)

        if hasattr(cls, 'OFP_VERSIONS') and cls.OFP_VERSIONS is not None:
            ofproto_protocol.set_app_supported_versions(cls.OFP_VERSIONS)

        if app_name is not None:
            assert app_name not in self.applications
        app = cls(*args, **kwargs)
        register_app(app)
        assert app.name not in self.applications
        self.applications[app.name] = app
        return app

    def instantiate(self, cls, *args, **kwargs):
        app = self._instantiate(None, cls, *args, **kwargs)
        self._update_bricks()
        self._report_brick(app.name, app)
        return app

    def instantiate_apps(self, *args, **kwargs):
        for app_name, cls in self.applications_cls.items():
            self._instantiate(app_name, cls, *args, **kwargs)

        self._update_bricks()
        self.report_bricks()

        threads = []
        for app in self.applications.values():
            t = app.start()
            if t is not None:
                app.set_main_thread(t)
                threads.append(t)
        return threads

    @staticmethod
    def _close(app):
        close_method = getattr(app, 'close', None)
        if callable(close_method):
            close_method()

    def uninstantiate(self, name):
        app = self.applications.pop(name)
        unregister_app(app)
        for app_ in SERVICE_BRICKS.values():
            app_.unregister_observer_all_event(name)
        app.stop()
        self._close(app)
        events = app.events
        if not events.empty():
            app.logger.debug('%s events remains %d', app.name, events.qsize())

    def close(self):
        def close_all(close_dict):
            for app in close_dict.values():
                self._close(app)
            close_dict.clear()

        # This semaphore prevents parallel execution of this function,
        # as run_apps's finally clause starts another close() call.
        with self.close_sem:
            for app_name in list(self.applications.keys()):
                self.uninstantiate(app_name)
            assert not self.applications
            close_all(self.contexts)



# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2016 Fumihiko Kakuma <kakuma at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from os_ken.tests.integrated.common import docker_base as ctn_base
from . import base


class BgpSpeakerBasicTest(base.BgpSpeakerTestBase):
    def setUp(self):
        super(BgpSpeakerBasicTest, self).setUp()
        self.r1.stop_os_kenbgp(retry=True)
        self.r1.start_os_kenbgp(retry=True)

    def test_check_neighbor_established(self):
        neighbor_state = ctn_base.BGP_FSM_IDLE
        for _ in range(0, self.checktime):
            neighbor_state = self.q1.get_neighbor_state(self.r1)
            if neighbor_state == ctn_base.BGP_FSM_ESTABLISHED:
                break
            time.sleep(1)
        self.assertEqual(neighbor_state, ctn_base.BGP_FSM_ESTABLISHED)

    def test_check_rib_nexthop(self):
        neighbor_state = ctn_base.BGP_FSM_IDLE
        for _ in range(0, self.checktime):
            neighbor_state = self.q1.get_neighbor_state(self.r1)
            if neighbor_state == ctn_base.BGP_FSM_ESTABLISHED:
                break
            time.sleep(1)
        self.assertEqual(neighbor_state, ctn_base.BGP_FSM_ESTABLISHED)
        rib = self.q1.get_global_rib(prefix='10.10.0.0/28')
        self.assertEqual(self.r1_ip, rib[0]['nexthop'])



# Copyright (C) 2011, 2012 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2011 Isaku Yamahata <yamahata at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct


def msg_pack_into(fmt, buf, offset, *args):
    needed_len = offset + struct.calcsize(fmt)
    if len(buf) < needed_len:
        buf += bytearray(needed_len - len(buf))

    struct.pack_into(fmt, buf, offset, *args)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
VRRP manager that manages VRRP router instances
VRRPManager creates/deletes VRRPRouter, VRRPInterfaceMonitor
dynamically as requested.

Usage example
osken-manager --verbose \
    os_ken.services.protocols.vrrp.manager \
    os_ken.services.protocols.vrrp.dumper
"""

import time

from os_ken.base import app_manager
from os_ken.controller import handler
from os_ken.lib import hub
from os_ken.services.protocols.vrrp import event as vrrp_event
from os_ken.services.protocols.vrrp import monitor as vrrp_monitor
from os_ken.services.protocols.vrrp import router as vrrp_router


class VRRPInstance(object):
    def __init__(self, name, monitor_name, config, interface):
        super(VRRPInstance, self).__init__()
        self.name = name                        # vrrp_router.name
        self.monitor_name = monitor_name        # interface_monitor.name
        self.config = config
        self.interface = interface
        self.state = None

    def state_changed(self, new_state):
        self.state = new_state


class VRRPManager(app_manager.OSKenApp):
    @staticmethod
    def _instance_name(interface, vrid, is_ipv6):
        ip_version = 'ipv6' if is_ipv6 else 'ipv4'
        return 'VRRP-Router-%s-%d-%s' % (str(interface), vrid, ip_version)

    def __init__(self, *args, **kwargs):
        super(VRRPManager, self).__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs
        self.name = vrrp_event.VRRP_MANAGER_NAME
        self._instances = {}    # name -> VRRPInstance
        self.shutdown = hub.Queue()

    def start(self):
        t = hub.spawn(self._shutdown_loop)
        super(VRRPManager, self).start()
        return t

    @handler.set_ev_cls(vrrp_event.EventVRRPConfigRequest)
    def config_request_handler(self, ev):
        config = ev.config
        interface = ev.interface
        name = self._instance_name(interface, config.vrid, config.is_ipv6)
        if name in self._instances:
            rep = vrrp_event.EventVRRPConfigReply(None, interface, config)
            self.reply_to_request(ev, rep)
            return

        statistics = VRRPStatistics(name, config.resource_id,
                                    config.statistics_interval)
        monitor = vrrp_monitor.VRRPInterfaceMonitor.factory(
            interface, config, name, statistics, *self._args, **self._kwargs)
        router = vrrp_router.VRRPRouter.factory(name, monitor.name, interface,
                                                config, statistics,
                                                *self._args, **self._kwargs)
        # Event piping
        #  vrrp_router -> vrrp_manager
        #    EventVRRPStateChanged to vrrp_manager is handled by framework
        #  vrrp_manager -> vrrp_rouer
        self.register_observer(vrrp_event.EventVRRPShutdownRequest,
                               router.name)
        #  vrrp_router -> vrrp_monitor
        router.register_observer(vrrp_event.EventVRRPStateChanged,
                                 monitor.name)
        router.register_observer(vrrp_event.EventVRRPTransmitRequest,
                                 monitor.name)
        #  vrrp_interface_monitor -> vrrp_router
        monitor.register_observer(vrrp_event.EventVRRPReceived, router.name)

        instance = VRRPInstance(name, monitor.name, config, interface)
        self._instances[name] = instance
        # self.logger.debug('report_bricks')
        # app_manager.AppManager.get_instance().report_bricks()   # debug
        monitor.start()
        router.start()

        rep = vrrp_event.EventVRRPConfigReply(instance.name, interface, config)
        self.reply_to_request(ev, rep)

    def _proxy_event(self, ev):
        name = ev.instance_name
        instance = self._instances.get(name, None)
        if not instance:
            self.logger.info('unknown vrrp router %s', name)
            return
        self.send_event(instance.name, ev)

    @handler.set_ev_cls(vrrp_event.EventVRRPShutdownRequest)
    def shutdown_request_handler(self, ev):
        self._proxy_event(ev)

    @handler.set_ev_cls(vrrp_event.EventVRRPConfigChangeRequest)
    def config_change_request_handler(self, ev):
        self._proxy_event(ev)

    @handler.set_ev_cls(vrrp_event.EventVRRPStateChanged)
    def state_change_handler(self, ev):
        instance = self._instances.get(ev.instance_name, None)
        assert instance is not None
        instance.state_changed(ev.new_state)
        if ev.old_state and ev.new_state == vrrp_event.VRRP_STATE_INITIALIZE:
            self.shutdown.put(instance)

    def _shutdown_loop(self):
        app_mgr = app_manager.AppManager.get_instance()
        while self.is_active or not self.shutdown.empty():
            instance = self.shutdown.get()
            app_mgr.uninstantiate(instance.name)
            app_mgr.uninstantiate(instance.monitor_name)
            del self._instances[instance.name]

    @handler.set_ev_cls(vrrp_event.EventVRRPListRequest)
    def list_request_handler(self, ev):
        instance_name = ev.instance_name
        if instance_name is None:
            instance_list = [vrrp_event.VRRPInstance(
                instance.name, instance.monitor_name,
                instance.config, instance.interface, instance.state)
                for instance in self._instances.values()]
        else:
            instance = self._instances.get(instance_name, None)
            if instance is None:
                instance_list = []
            else:
                instance_list = [vrrp_event.VRRPInstance(
                    instance_name, instance.monitor_name,
                    instance.config, instance.interface, instance.state)]

        vrrp_list = vrrp_event.EventVRRPListReply(instance_list)
        self.reply_to_request(ev, vrrp_list)


class VRRPStatistics(object):
    def __init__(self, name, resource_id, statistics_interval):
        self.name = name
        self.resource_id = resource_id
        self.statistics_interval = statistics_interval
        self.tx_vrrp_packets = 0
        self.rx_vrrp_packets = 0
        self.rx_vrrp_zero_prio_packets = 0
        self.tx_vrrp_zero_prio_packets = 0
        self.rx_vrrp_invalid_packets = 0
        self.rx_vrrp_bad_auth = 0
        self.idle_to_master_transitions = 0
        self.idle_to_backup_transitions = 0
        self.backup_to_master_transitions = 0
        self.master_to_backup_transitions = 0

    def get_stats(self):
        ts = time.strftime("%Y-%m-%dT%H:%M:%S")
        stats_dict = dict(
            timestamp=ts,
            resource_id=self.resource_id,
            tx_vrrp_packets=self.tx_vrrp_packets,
            rx_vrrp_packets=self.rx_vrrp_packets,
            rx_vrrp_zero_prio_packets=self.rx_vrrp_zero_prio_packets,
            tx_vrrp_zero_prio_packets=self.tx_vrrp_zero_prio_packets,
            rx_vrrp_invalid_packets=self.rx_vrrp_invalid_packets,
            rx_vrrp_bad_auth=self.rx_vrrp_bad_auth,
            idle_to_master_transitions=self.idle_to_master_transitions,
            idle_to_backup_transitions=self.idle_to_backup_transitions,
            backup_to_master_transitions=self.backup_to_master_transitions,
            master_to_backup_transitions=self.master_to_backup_transitions
        )
        return stats_dict



from os_ken.services.protocols.bgp.operator.views.base import \
    create_dict_view_class
from os_ken.services.protocols.bgp.operator.views.base import OperatorDetailView
from os_ken.services.protocols.bgp.operator.views import fields


class ImportmapManagerDetailView(OperatorDetailView):
    importmaps = fields.RelatedDictViewField(
        '_import_maps_by_name',
        'os_ken.services.protocols.bgp.operator.views.other.ImportmapDictView'
    )


class ImportmapDetailView(OperatorDetailView):
    nlri = fields.OptionalDataField('_nlri')
    rt = fields.OptionalDataField('_rt')

    def encode(self):
        ret = {}
        nlri = self.get_field('nlri')
        if nlri is not None:
            ret.update({'nlri': nlri})

        rt = self.get_field('rt')
        if rt is not None:
            ret.update({'rt': rt})

        return ret


ImportmapDictView = create_dict_view_class(
    ImportmapDetailView,
    'ImportmapDictView'
)



from os_ken.services.protocols.bgp.rtconf.base import ConfWithStats
from os_ken.services.protocols.bgp.rtconf.common import CommonConfListener
from os_ken.services.protocols.bgp.rtconf.neighbors import NeighborsConfListener
from os_ken.services.protocols.bgp.rtconf import vrfs
from os_ken.services.protocols.bgp.rtconf.vrfs import VrfConf
from os_ken.services.protocols.bgp.rtconf.vrfs import VrfsConfListener

import logging

LOG = logging.getLogger('bgpspeaker.core_managers.table_mixin')


class ConfigurationManager(CommonConfListener, VrfsConfListener,
                           NeighborsConfListener):
    def __init__(self, core_service, common_conf, vrfs_conf, neighbors_conf):
        self._signal_bus = core_service.signal_bus
        self._common_config = common_conf
        self._peer_manager = core_service.peer_manager
        self._table_manager = core_service.table_manager
        self._rt_manager = core_service.rt_manager
        CommonConfListener.__init__(self, common_conf)
        VrfsConfListener.__init__(self, vrfs_conf)
        NeighborsConfListener.__init__(self, neighbors_conf)

    def on_update_common_conf(self, evt):
        raise NotImplementedError()

    def on_add_neighbor_conf(self, evt):
        neigh_conf = evt.value
        self._peer_manager.add_peer(neigh_conf, self._common_config)

    def on_remove_neighbor_conf(self, evt):
        neigh_conf = evt.value
        self._peer_manager.remove_peer(neigh_conf)

    def on_chg_vrf_conf(self, evt):
        evt_value = evt.value
        vrf_conf = evt.src
        new_imp_rts, removed_imp_rts, import_maps, re_export, re_import = \
            evt_value
        route_family = vrf_conf.route_family
        vrf_table = self._table_manager.get_vrf_table(
            vrf_conf.route_dist, route_family
        )
        assert vrf_table

        # If we have new import RTs we have to update RTC table and make route
        # refresh request to peers not participating in RT address-family
        self._table_manager.update_vrf_table_links(
            vrf_table, new_imp_rts, removed_imp_rts
        )

        # If other properties of VRF changed we re-install local paths.
        if re_export:
            self._table_manager.re_install_net_ctrl_paths(vrf_table)

        # We have to withdraw paths that do not have any RT that are or
        # interest
        vrf_table.clean_uninteresting_paths()
        if import_maps is not None:
            vrf_table.init_import_maps(import_maps)
            changed_dests = vrf_table.apply_import_maps()
            for dest in changed_dests:
                self._signal_bus.dest_changed(dest)

        # import new rts
        if re_import:
            LOG.debug(
                "RE-importing prefixes from VPN table to VRF %r", vrf_table
            )
            self._table_manager.import_all_vpn_paths_to_vrf(vrf_table)
        else:
            self._table_manager.import_all_vpn_paths_to_vrf(
                vrf_table, new_imp_rts
            )

        # Update local/global RT NLRIs
        self._rt_manager.update_local_rt_nlris()

    def on_remove_vrf_conf(self, evt):
        """Removes VRF table associated with given `vrf_conf`.

        Cleans up other links to this table as well.
        """
        vrf_conf = evt.value
        # Detach VrfConf change listener.
        vrf_conf.remove_listener(VrfConf.VRF_CHG_EVT, self.on_chg_vrf_conf)

        self._table_manager.remove_vrf_by_vrf_conf(vrf_conf)

        # Update local RT NLRIs
        self._rt_manager.update_local_rt_nlris()

        self._signal_bus.vrf_removed(vrf_conf.route_dist)

        # Remove AttributeMaps under the removed vrf
        rd = vrf_conf.route_dist
        rf = vrf_conf.route_family
        peers = self._peer_manager.iterpeers
        for peer in peers:
            key = ':'.join([rd, rf])
            peer.attribute_maps.pop(key, None)

    def on_add_vrf_conf(self, evt):
        """Event handler for new VrfConf.

        Creates a VrfTable to store routing information related to new Vrf.
        Also arranges for related paths to be imported to this VrfTable.
        """
        vrf_conf = evt.value
        route_family = vrf_conf.route_family
        assert route_family in vrfs.SUPPORTED_VRF_RF
        # Create VRF table with given configuration.
        vrf_table = self._table_manager.create_and_link_vrf_table(vrf_conf)

        # Attach VrfConf change listeners.
        vrf_conf.add_listener(ConfWithStats.UPDATE_STATS_LOG_ENABLED_EVT,
                              self.on_stats_config_change)
        vrf_conf.add_listener(ConfWithStats.UPDATE_STATS_TIME_EVT,
                              self.on_stats_config_change)
        vrf_conf.add_listener(VrfConf.VRF_CHG_EVT, self.on_chg_vrf_conf)

        # Import paths from VPN table that match this VRF/VPN.
        self._table_manager.import_all_vpn_paths_to_vrf(vrf_table)

        # Update local RT NLRIs
        self._rt_manager.update_local_rt_nlris()
        self._signal_bus.vrf_added(vrf_conf)

    def on_stats_config_change(self, evt):
        vrf_conf = evt.src
        self._signal_bus.stats_config_changed(vrf_conf)



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import six
import struct
from . import packet_base
from . import arp
from . import ipv4
from . import ipv6
from . import lldp
from . import slow
from . import llc
from . import pbb
from . import cfm
from . import ether_types as ether


@six.add_metaclass(abc.ABCMeta)
class _vlan(packet_base.PacketBase):
    _PACK_STR = "!HH"
    _MIN_LEN = struct.calcsize(_PACK_STR)

    @abc.abstractmethod
    def __init__(self, pcp, cfi, vid, ethertype):
        super(_vlan, self).__init__()
        self.pcp = pcp
        self.cfi = cfi
        self.vid = vid
        self.ethertype = ethertype

    @classmethod
    def parser(cls, buf):
        tci, ethertype = struct.unpack_from(cls._PACK_STR, buf)
        pcp = tci >> 13
        cfi = (tci >> 12) & 1
        vid = tci & ((1 << 12) - 1)
        return (cls(pcp, cfi, vid, ethertype),
                vlan.get_packet_type(ethertype), buf[vlan._MIN_LEN:])

    def serialize(self, payload, prev):
        tci = self.pcp << 13 | self.cfi << 12 | self.vid
        return struct.pack(vlan._PACK_STR, tci, self.ethertype)


class vlan(_vlan):
    """VLAN (IEEE 802.1Q) header encoder/decoder class.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    ============== ====================
    Attribute      Description
    ============== ====================
    pcp            Priority Code Point
    cfi            Canonical Format Indicator
    vid            VLAN Identifier
    ethertype      EtherType
    ============== ====================
    """

    def __init__(self, pcp=0, cfi=0, vid=0, ethertype=ether.ETH_TYPE_IP):
        super(vlan, self).__init__(pcp, cfi, vid, ethertype)

    @classmethod
    def get_packet_type(cls, type_):
        """Override method for the Length/Type field (self.ethertype).
        The Length/Type field means Length or Type interpretation,
        same as ethernet IEEE802.3.
        If the value of Length/Type field is less than or equal to
        1500 decimal(05DC hexadecimal), it means Length interpretation
        and be passed to the LLC sublayer."""
        if type_ <= ether.ETH_TYPE_IEEE802_3:
            type_ = ether.ETH_TYPE_IEEE802_3
        return cls._TYPES.get(type_)


class svlan(_vlan):
    """S-VLAN (IEEE 802.1ad) header encoder/decoder class.


    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== ====================
    Attribute      Description
    ============== ====================
    pcp            Priority Code Point
    cfi            Canonical Format Indicator.
                   In a case to be used as B-TAG,
                   this field means DEI(Drop Eligible Indication).
    vid            VLAN Identifier
    ethertype      EtherType
    ============== ====================
    """

    def __init__(self, pcp=0, cfi=0, vid=0, ethertype=ether.ETH_TYPE_8021Q):
        super(svlan, self).__init__(pcp, cfi, vid, ethertype)

    @classmethod
    def get_packet_type(cls, type_):
        return cls._TYPES.get(type_)


vlan.register_packet_type(arp.arp, ether.ETH_TYPE_ARP)
vlan.register_packet_type(ipv4.ipv4, ether.ETH_TYPE_IP)
vlan.register_packet_type(ipv6.ipv6, ether.ETH_TYPE_IPV6)
vlan.register_packet_type(lldp.lldp, ether.ETH_TYPE_LLDP)
vlan.register_packet_type(slow.slow, ether.ETH_TYPE_SLOW)
vlan.register_packet_type(llc.llc, ether.ETH_TYPE_IEEE802_3)
vlan.register_packet_type(cfm.cfm, ether.ETH_TYPE_CFM)

svlan.register_packet_type(vlan, ether.ETH_TYPE_8021Q)
svlan.register_packet_type(pbb.itag, ether.ETH_TYPE_8021AH)
